# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module provides abstract base classes which represent the quantum state
returned by a simulator backend via :class:`.Engine`.
"""
import abc
import string
from itertools import chain
from copy import copy

import numpy as np
from scipy.linalg import block_diag
from scipy.stats import multivariate_normal
from scipy.special import factorial
from scipy.integrate import simps

import strawberryfields as sf
from .shared_ops import rotation_matrix as _R
from .shared_ops import changebasis

indices = string.ascii_lowercase


class BaseState(abc.ABC):
    r"""Abstract base class for the representation of quantum states."""
    EQ_TOLERANCE = 1e-10

    def __init__(self, num_modes, mode_names=None):
        self._modes = num_modes
        self._hbar = sf.hbar  # always use the global frontend hbar value for state objects
        self._data = None
        self._pure = None

        if mode_names is None:
            self._modemap = {i: "mode {}".format(i) for i in range(num_modes)}
        else:
            self._modemap = {i: "{}".format(j) for i, j in zip(range(num_modes), mode_names)}

        self._str = "<BaseState: num_modes={}, pure={}, hbar={}>".format(
            self.num_modes, self._pure, self._hbar
        )

    def __str__(self):
        return self._str

    def __repr__(self):
        return self._str

    @property
    def data(self):
        r"""Returns the underlying numerical (or symbolic) representation of the state.
        The form of this data differs for different backends."""
        return self._data

    @property
    def hbar(self):
        r"""Returns the value of :math:`\hbar` used in the generation of the state.

        The value of :math:`\hbar` is a convention chosen in the definition of
        :math:`\x` and :math:`\p`. See :ref:`opcon` for more details.

        Returns:
            float: :math:`\hbar` value.
        """
        return self._hbar

    @property
    def is_pure(self):
        r"""Checks whether the state is a pure state.

        Returns:
            bool: True if and only if the state is pure.
        """
        return self._pure

    @property
    def num_modes(self):
        r"""Gets the number of modes that the state represents.

        Returns:
            int: the number of modes in the state
        """
        return self._modes

    @property
    def mode_names(self):
        r"""Returns a dictionary mapping the mode index to mode names.

        The mode names are determined from the initialization argument
        ``mode_names``. If these were not supplied, the names are generated automatically based
        on the mode indices.

        Returns:
            dict: dictionary of the form ``{i:"mode name",...}``
        """
        return self._modemap

    @property
    def mode_indices(self):
        r"""Returns a dictionary mapping the mode names to mode indices.

        The mode names are determined from the initialization argument
        ``mode_names``. If these were not supplied, the names are generated automatically based
        on the mode indices.

        Returns:
            dict: dictionary of the form ``{"mode name":i,...}``
        """
        return {v: k for k, v in self._modemap.items()}

    @abc.abstractmethod
    def reduced_dm(self, modes, **kwargs):
        r"""Returns a reduced density matrix in the Fock basis.

        Args:
            modes (int or Sequence[int]): specifies the mode(s) to return the reduced density matrix for.
            **kwargs:

                  * **cutoff** (*int*): (default 10) specifies where to truncate the returned density matrix.
                    Note that the cutoff argument only applies for Gaussian representation;
                    states represented in the Fock basis will use their own internal cutoff dimension.

        Returns:
            array: the reduced density matrix for the specified modes
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fock_prob(self, n, **kwargs):
        r"""Probability of a particular Fock basis state.

        Computes the probability :math:`|\braket{\vec{n}|\psi}|^2` of measuring
        the given multi-mode Fock state based on the state :math:`\ket{\psi}`.

        .. warning::

            Computing the Fock probabilities of states has exponential scaling
            in the Gaussian representation (for example states output by a
            Gaussian backend as a :class:`~.BaseGaussianState`).
            This shouldn't affect small-scale problems, where only a few Fock
            basis state probabilities need to be calculated, but will become
            evident in larger scale problems.

        Args:
            n (Sequence[int]): the Fock state :math:`\ket{\vec{n}}` that we want to measure the probability of
            **kwargs:

                  * **cutoff** (*int*): (default 10) specifies the fock basis truncation when calculating
                    of the fock basis probabilities.
                    Note that the cutoff argument only applies for Gaussian representation;
                    states represented in the Fock basis will use their own internal cutoff dimension.

        Returns:
            float: measurement probability
        """
        raise NotImplementedError

    @abc.abstractmethod
    def mean_photon(self, mode, **kwargs):
        """Returns the mean photon number of a particular mode.

        Args:
            mode (int): specifies the mode
            **kwargs:

                  * **cutoff** (*int*): (default 10) Fock basis trunction for calculation of
                    mean photon number.
                    Note that the cutoff argument only applies for Gaussian representation;
                    states represented in the Fock basis will use their own internal cutoff dimension.

        Returns:
            tuple: the mean photon number and variance
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fidelity(self, other_state, mode, **kwargs):
        r"""Fidelity of the reduced state in the specified mode with a user supplied state.
        Note that this method only supports single-mode states.

        Args:
            other_state: a pure state vector array represented in the Fock basis (for Fock backends)
                or a Sequence ``(mu, cov)`` containing the means and covariance matrix (for Gaussian backends)

        Returns:
            The fidelity of the circuit state with ``other_state``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fidelity_vacuum(self, **kwargs):
        """The fidelity of the state with the vacuum state.

        Returns:
            float: the fidelity of the state with the vacuum
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fidelity_coherent(self, alpha_list, **kwargs):
        r"""The fidelity of the state with a product of coherent states.

        The fidelity is defined by

        .. math:: \bra{\vec{\alpha}}\rho\ket{\vec{\alpha}}

        Args:
            alpha_list (Sequence[complex]): list of coherent state parameters, one for each mode

        Returns:
            float: the fidelity value
        """
        raise NotImplementedError

    @abc.abstractmethod
    def wigner(self, mode, xvec, pvec):
        r"""Calculates the discretized Wigner function of the specified mode.

        Args:
            mode (int): the mode to calculate the Wigner function for
            xvec (array): array of discretized :math:`x` quadrature values
            pvec (array): array of discretized :math:`p` quadrature values

        Returns:
            array: 2D array of size [len(xvec), len(pvec)], containing reduced Wigner function
            values for specified x and p values.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def quad_expectation(self, mode, phi=0, **kwargs):
        r"""The :math:`\x_{\phi}` operator expectation values and variance for the specified mode.

        The :math:`\x_{\phi}` operator is defined as follows,

        .. math:: \x_{\phi} = \cos\phi~\x + \sin\phi~\p

        with corresponding expectation value

        .. math:: \bar{x_{\phi}}=\langle x_{\phi}\rangle = \text{Tr}(\x_{\phi}\rho_{mode})

        and variance

        .. math:: \Delta x_{\phi}^2 = \langle x_{\phi}^2\rangle - \braket{x_{\phi}}^2

        Args:
            mode (int): the requested mode
            phi (float): quadrature angle, clockwise from the positive :math:`x` axis.

                * :math:`\phi=0` corresponds to the :math:`x` expectation and variance (default)
                * :math:`\phi=\pi/2` corresponds to the :math:`p` expectation and variance

        Returns:
            tuple (float, float): expectation value and variance
        """
        raise NotImplementedError

    @abc.abstractmethod
    def poly_quad_expectation(self, A, d=None, k=0, phi=0, **kwargs):
        r"""The multi-mode expectation values and variance of arbitrary 2nd order polynomials
        of quadrature operators.

        An arbitrary 2nd order polynomial of quadrature operators over $N$ modes can always
        be written in the following form:

        .. math:: P(\mathbf{r}) = \mathbf{r}^T A\mathbf{r} + \mathbf{r}^T \mathbf{d} + k I

        where:

        * :math:`A\in\mathbb{R}^{2N\times 2N}` is a symmetric matrix
          representing the quadratic coefficients,
        * :math:`\mathbf{d}\in\mathbb{R}^{2N}` is a real vector representing
          the linear coefficients,
        * :math:`k\in\mathbb{R}` represents the constant term, and
        * :math:`\mathbf{r} = (\x_1,\dots,\x_N,\p_1,\dots,\p_N)` is the vector
          of quadrature operators in :math:`xp`-ordering.

        This method returns the expectation value of this second-order polynomial,

        .. math:: \langle P(\mathbf{r})\rangle,

        as well as the variance

        .. math:: \Delta P(\mathbf{r})^2 = \braket{P(\mathbf{r})^2} - \braket{P(\mathbf{r})}^2

        Args:
            A (array): a real symmetric 2Nx2N NumPy array, representing the quadratic
                coefficients of the second order quadrature polynomial.
            d (array): a real length-2N NumPy array, representing the linear
                coefficients of the second order quadrature polynomial. Defaults to the zero vector.
            k (float): the constant term. Default 0.
            phi (float): quadrature angle, clockwise from the positive :math:`x` axis. If provided,
                the vectori of quadrature operators :math:`\mathbf{r}` is first rotated
                by angle :math:`\phi` in the phase space.


        Returns:
            tuple (float, float): expectation value and variance
        """
        raise NotImplementedError

    def p_quad_values(self, mode, xvec, pvec):

        r"""Calculates the discretized p-quadrature probability distribution of the specified mode.

        Args:
            mode (int): the mode to calculate the p-quadrature probability values of
            xvec (array): array of discretized :math:`x` quadrature values
            pvec (array): array of discretized :math:`p` quadrature values

        Returns:
            array: 1D array of size len(pvec), containing reduced p-quadrature
            probability values for a specified range of x and p.
        """

        W = self.wigner(mode, xvec, pvec)
        y = []
        for i in range(0, len(pvec)):
            res = simps(W[i, : len(xvec)], xvec)
            y.append(res)
        return np.array(y)

    def x_quad_values(self, mode, xvec, pvec):

        r"""Calculates the discretized x-quadrature probability distribution of the specified mode.

        Args:
            mode (int): the mode to calculate the x-quadrature probability values of
            xvec (array): array of discretized :math:`x` quadrature values
            pvec (array): array of discretized :math:`p` quadrature values

        Returns:
            array: 1D array of size len(xvec), containing reduced x-quadrature
            probability values for a specified range of x and p.
        """

        W = self.wigner(mode, xvec, pvec)
        y = []
        for i in range(0, len(xvec)):
            res = simps(W[: len(pvec), i], pvec)
            y.append(res)
        return np.array(y)


class BaseFockState(BaseState):
    r"""Class for the representation of quantum states in the Fock basis.

    Args:
        state_data (array): the state representation in the Fock basis
        num_modes (int): the number of modes in the state
        pure (bool): True if the state is a pure state, false if the state is mixed
        cutoff_dim (int): the Fock basis truncation size
        mode_names (Sequence): (optional) this argument contains a list providing mode names
            for each mode in the state
    """

    def __init__(self, state_data, num_modes, pure, cutoff_dim, mode_names=None):
        # pylint: disable=too-many-arguments

        super().__init__(num_modes, mode_names)

        self._data = state_data
        self._cutoff = cutoff_dim
        self._pure = pure
        self._basis = "fock"

        self._str = "<FockState: num_modes={}, cutoff={}, pure={}, hbar={}>".format(
            self.num_modes, self._cutoff, self._pure, self._hbar
        )

    def __eq__(self, other):
        """Equality operator for BaseFockState.

        Returns True if other BaseFockState is close to self.
        This is done by comparing the dm attribute - if within
        the EQ_TOLERANCE, True is returned.

        Args:
            other (BaseFockState): BaseFockState to compare against.
        """
        if not isinstance(other, type(self)):
            return False

        if self.num_modes != other.num_modes:
            return False

        if self.data.shape != other.data.shape:
            return False

        if np.allclose(self.dm(), other.dm(), atol=self.EQ_TOLERANCE, rtol=0):
            return True

        return False

    @property
    def cutoff_dim(self):
        r"""The numerical truncation of the Fock space used by the underlying state.
        Note that a cutoff of D corresponds to the Fock states :math:`\{|0\rangle,\dots,|D-1\rangle\}`

        Returns:
            int: the cutoff dimension
        """
        return self._cutoff

    def ket(self, **kwargs):
        r"""The numerical state vector for the quantum state.
        Note that if the state is mixed, this method returns None.

        Returns:
            array/None: the numerical state vector. Returns None if the state is mixed.
        """
        # pylint: disable=unused-argument
        if self._pure:
            return self.data

        return None  # pragma: no cover

    def dm(self, **kwargs):
        r"""The numerical density matrix for the quantum state.

        Returns:
            array: the numerical density matrix in the Fock basis
        """
        # pylint: disable=unused-argument
        if self._pure:
            left_str = [indices[i] for i in range(0, 2 * self._modes, 2)]
            right_str = [indices[i] for i in range(1, 2 * self._modes, 2)]
            out_str = [indices[: 2 * self._modes]]
            einstr = "".join(left_str + [","] + right_str + ["->"] + out_str)
            rho = np.einsum(einstr, self.ket(), self.ket().conj())
            return rho

        return self.data

    def trace(self, **kwargs):
        r"""Trace of the density operator corresponding to the state.

        For pure states the trace corresponds to the squared norm of the ket vector.

        For physical states this should always be 1, any deviations from this value are due
        to numerical errors and Hilbert space truncation artefacts.

        Returns:
          float: trace of the state
        """
        # pylint: disable=unused-argument
        if self.is_pure:
            return np.vdot(self.ket(), self.ket()).real  # <s|s>

        # need some extra steps to trace over multimode matrices
        eqn_indices = [
            [indices[idx]] * 2 for idx in range(self._modes)
        ]  # doubled indices [['i','i'],['j','j'], ... ]
        eqn = "".join(
            chain.from_iterable(eqn_indices)
        )  # flatten indices into a single string 'iijj...'
        return np.einsum(eqn, self.dm()).real

    def all_fock_probs(self, **kwargs):
        r"""Probabilities of all possible Fock basis states for the current circuit state.

        For example, in the case of 3 modes, this method allows the Fock state probability
        :math:`|\braketD{0,2,3}{\psi}|^2` to be returned via

        .. code-block:: python

            probs = state.all_fock_probs()
            probs[0,2,3]

        Returns:
            array: array of dimension :math:`\underbrace{D\times D\times D\cdots\times D}_{\text{num modes}}`
                containing the Fock state probabilities, where :math:`D` is the Fock basis cutoff truncation
        """
        # pylint: disable=unused-argument
        if self._pure:
            s = np.ravel(self.ket())  # into 1D array
            return np.reshape((s * s.conj()).real, [self._cutoff] * self._modes)

        s = self.dm()
        num_axes = len(s.shape)
        evens = [k for k in range(0, num_axes, 2)]
        odds = [k for k in range(1, num_axes, 2)]
        flat_size = np.prod([s.shape[k] for k in range(0, num_axes, 2)])
        transpose_list = evens + odds
        probs = np.diag(np.reshape(np.transpose(s, transpose_list), [flat_size, flat_size])).real

        return np.reshape(probs, [self._cutoff] * self._modes)

    # =====================================================
    # the following methods are overwritten from BaseState

    def reduced_dm(self, modes, **kwargs):
        # pylint: disable=unused-argument
        if modes == list(range(self._modes)):
            # reduced state is full state
            return self.dm()  # pragma: no cover

        if isinstance(modes, int):
            modes = [modes]
        if modes != sorted(modes):
            raise ValueError("The specified modes cannot be duplicated.")

        if len(modes) > self._modes:
            raise ValueError(
                "The number of specified modes cannot " "be larger than the number of subsystems."
            )

        # reduce rho down to specified subsystems
        keep_indices = indices[: 2 * len(modes)]
        trace_indices = indices[2 * len(modes) : len(modes) + self._modes]

        ind = [i * 2 for i in trace_indices]
        ctr = 0

        for m in range(self._modes):
            if m in modes:
                ind.insert(m, keep_indices[2 * ctr : 2 * (ctr + 1)])
                ctr += 1

        indStr = "".join(ind) + "->" + keep_indices
        return np.einsum(indStr, self.dm())

    def fock_prob(self, n, **kwargs):
        # pylint: disable=unused-argument
        if len(n) != self._modes:
            raise ValueError("List length should be equal to number of modes")

        elif max(n) >= self._cutoff:
            raise ValueError("Can't get distribution beyond truncation level")

        if self._pure:
            return np.abs(self.ket()[tuple(n)]) ** 2

        return self.dm()[tuple([n[i // 2] for i in range(len(n) * 2)])].real

    def mean_photon(self, mode, **kwargs):
        # pylint: disable=unused-argument
        n = np.arange(self._cutoff)
        probs = np.diagonal(self.reduced_dm(mode))
        mean = np.sum(n * probs).real
        var = np.sum(n ** 2 * probs).real - mean ** 2
        return mean, var

    def fidelity(self, other_state, mode, **kwargs):
        # pylint: disable=unused-argument
        max_indices = len(indices) // 2

        if self.num_modes > max_indices:
            raise Exception("fidelity method can only support up to {} modes".format(max_indices))

        left_indices = indices[:mode]
        eqn_left = "".join([i * 2 for i in left_indices])
        reduced_dm_indices = indices[mode : mode + 2]
        right_indices = indices[mode + 2 : self._modes + 1]
        eqn_right = "".join([i * 2 for i in right_indices])
        eqn = "".join([eqn_left, reduced_dm_indices, eqn_right]) + "->" + reduced_dm_indices
        rho_reduced = np.einsum(eqn, self.dm())

        return np.dot(np.conj(other_state), np.dot(rho_reduced, other_state)).real

    def fidelity_vacuum(self, **kwargs):
        # pylint: disable=unused-argument
        alpha = np.zeros(self._modes)
        return self.fidelity_coherent(alpha)

    def fidelity_coherent(self, alpha_list, **kwargs):
        # pylint: disable=too-many-locals,unused-argument
        if self.is_pure:
            mode_size = 1
            s = self.ket()
        else:
            mode_size = 2
            s = self.dm()

        if not hasattr(alpha_list, "__len__"):
            alpha_list = [alpha_list]  # pragma: no cover

        if len(alpha_list) != self._modes:
            raise ValueError("The number of alpha values must match the number of modes.")

        coh = lambda a, dim: np.array(
            [np.exp(-0.5 * np.abs(a) ** 2) * (a) ** n / np.sqrt(factorial(n)) for n in range(dim)]
        )

        if self._modes == 1:
            multi_cohs_vec = coh(alpha_list[0], self._cutoff)
        else:
            multi_cohs_list = [
                coh(alpha_list[idx], dim) for idx, dim in enumerate(s.shape[::mode_size])
            ]
            eqn = ",".join(indices[: self._modes]) + "->" + indices[: self._modes]
            multi_cohs_vec = np.einsum(
                eqn, *multi_cohs_list
            )  # tensor product of specified coherent states

        if self.is_pure:
            ovlap = np.vdot(multi_cohs_vec, s)
            return np.abs(ovlap) ** 2

        bra_indices = indices[: 2 * self._modes : 2]
        ket_indices = indices[1 : 2 * self._modes : 2]
        new_eqn_lhs = ",".join([bra_indices, ket_indices])
        new_eqn_rhs = "".join(bra_indices[idx] + ket_indices[idx] for idx in range(self._modes))
        outer_prod_eqn = new_eqn_lhs + "->" + new_eqn_rhs
        multi_coh_matrix = np.einsum(outer_prod_eqn, multi_cohs_vec, np.conj(multi_cohs_vec))

        return np.vdot(s, multi_coh_matrix).real

    def wigner(self, mode, xvec, pvec):
        r"""Calculates the discretized Wigner function of the specified mode.

        .. note::

            This code is a modified version of the 'iterative' method of the
            `wigner function provided in QuTiP <http://qutip.org/docs/4.0.2/apidoc/functions.html?highlight=wigner#qutip.wigner.wigner>`_,
            which is released under the BSD license, with the following
            copyright notice:

            Copyright (C) 2011 and later, P.D. Nation, J.R. Johansson,
            A.J.G. Pitchford, C. Granade, and A.L. Grimsmo. All rights reserved.

        Args:
            mode (int): the mode to calculate the Wigner function for
            xvec (array): array of discretized :math:`x` quadrature values
            pvec (array): array of discretized :math:`p` quadrature values

        Returns:
            array: 2D array of size [len(xvec), len(pvec)], containing reduced Wigner function
            values for specified x and p values.
        """
        rho = self.reduced_dm(mode)
        Q, P = np.meshgrid(xvec, pvec)
        A = (Q + P * 1.0j) / (2 * np.sqrt(self._hbar / 2))

        Wlist = np.array([np.zeros(np.shape(A), dtype=complex) for k in range(self._cutoff)])

        # Wigner function for |0><0|
        Wlist[0] = np.exp(-2.0 * np.abs(A) ** 2) / np.pi

        # W = rho(0,0)W(|0><0|)
        W = np.real(rho[0, 0]) * np.real(Wlist[0])

        for n in range(1, self._cutoff):
            Wlist[n] = (2.0 * A * Wlist[n - 1]) / np.sqrt(n)
            W += 2 * np.real(rho[0, n] * Wlist[n])

        for m in range(1, self._cutoff):
            temp = copy(Wlist[m])
            # Wlist[m] = Wigner function for |m><m|
            Wlist[m] = (2 * np.conj(A) * temp - np.sqrt(m) * Wlist[m - 1]) / np.sqrt(m)

            # W += rho(m,m)W(|m><m|)
            W += np.real(rho[m, m] * Wlist[m])

            for n in range(m + 1, self._cutoff):
                temp2 = (2 * A * Wlist[n - 1] - np.sqrt(m) * temp) / np.sqrt(n)
                temp = copy(Wlist[n])
                # Wlist[n] = Wigner function for |m><n|
                Wlist[n] = temp2

                # W += rho(m,n)W(|m><n|) + rho(n,m)W(|n><m|)
                W += 2 * np.real(rho[m, n] * Wlist[n])

        return W / (self._hbar)

    def quad_expectation(self, mode, phi=0, **kwargs):
        a = np.diag(np.sqrt(np.arange(1, self._cutoff + 5)), 1)
        x = np.sqrt(self._hbar / 2) * (a + a.T)
        p = -1j * np.sqrt(self._hbar / 2) * (a - a.T)

        xphi = np.cos(phi) * x + np.sin(phi) * p
        xphisq = np.dot(xphi, xphi)

        # truncate down
        xphi = xphi[: self._cutoff, : self._cutoff]
        xphisq = xphisq[: self._cutoff, : self._cutoff]

        rho = self.reduced_dm(mode)

        mean = np.trace(np.dot(xphi, rho)).real
        var = np.trace(np.dot(xphisq, rho)).real - mean ** 2

        return mean, var

    def poly_quad_expectation(self, A, d=None, k=0, phi=0, **kwargs):
        # pylint: disable=too-many-branches

        if A is None:
            A = np.zeros([2 * self._modes, 2 * self._modes])

        if A.shape != (2 * self._modes, 2 * self._modes):
            raise ValueError("Matrix of quadratic coefficients A must be of size 2Nx2N.")

        if not np.allclose(A.T, A):
            raise ValueError("Matrix of quadratic coefficients A must be symmetric.")

        if d is None:
            linear_coeff = np.zeros([2 * self._modes])
        else:
            linear_coeff = d.copy()
            linear_coeff[self._modes :] = -d[self._modes :]

        if linear_coeff.shape != (2 * self._modes,):
            raise ValueError("Vector of linear coefficients d must be of length 2N.")

        # expand the cutoff dimension in approximating the x and p
        # operators in the Fock basis, to reduce numerical inaccuracy.
        worksize = 1
        dim = self._cutoff + worksize

        # construct the x and p operators
        a = np.diag(np.sqrt(np.arange(1, dim)), 1)
        x_ = np.sqrt(self._hbar / 2) * (a + a.T)
        p_ = -1j * np.sqrt(self._hbar / 2) * (a - a.T)

        if phi != 0:
            # rotate the quadrature operators
            x = np.cos(phi) * x_ - np.sin(phi) * p_
            p = np.sin(phi) * x_ + np.cos(phi) * p_
        else:
            x = x_
            p = p_

        def expand_dims(op, n, modes):
            """Expand quadrature operator to act on nth mode"""
            I = np.identity(dim)
            allowed_indices = zip(indices[: 2 * modes : 2], indices[1 : 2 * modes : 2])
            ind = ",".join(a + b for a, b in allowed_indices)
            ops = [I] * n + [op] + [I] * (modes - n - 1)
            # the einsum 'ij,kl,mn->ijklmn' (for 3 modes)
            return np.einsum(ind, *ops)

        # determine modes with quadratic expectation values
        nonzero = np.concatenate(
            [np.mod(A.nonzero()[0], self._modes), np.mod(linear_coeff.nonzero()[0], self._modes)]
        )
        ex_modes = list(set(nonzero))
        num_modes = len(ex_modes)

        if not ex_modes:
            # only a constant term was provided
            return k, 0.0

        # There are non-zero elements of A and/or d
        # therefore there are quadratic and/or linear terms.
        # find the reduced density matrix
        rho = self.reduced_dm(modes=ex_modes)

        # generate vector of quadrature operators
        # this array will have shape [2*num_modes] + [dim]*(2*num_modes)
        r = np.empty([2 * num_modes] + [dim] * (2 * num_modes), dtype=np.complex128)
        for n in range(num_modes):
            r[n] = expand_dims(x, n, num_modes)
            r[num_modes + n] = expand_dims(p, n, num_modes)

        # reduce the size of A so that we only consider modes
        # which we need to calculate the expectation value for
        rows = ex_modes + [i + self._modes for i in ex_modes]
        quad_coeffs = A[:, rows][rows]
        quad_coeffs[num_modes:, :num_modes] = -quad_coeffs[num_modes:, :num_modes]
        quad_coeffs[:num_modes, num_modes:] = -quad_coeffs[:num_modes, num_modes:]

        # Compute the polynomial
        #
        # For 3 modes, this gives the einsum (with brackets denoting modes):
        # 'a(bc)(de)(fg),a(ch)(ei)(gj)->(bh)(di)(fj)' applied to r, A@r
        #
        # a corresponds to the index in the vector of quadrature operators
        # r = (x_1,...,x_n,p_1,...,p_n), and the remaining indices ijklmn
        # are the elements of the operator acting on a 3 mode density matrix.
        #
        # So, in effect, matrix of quadratic coefficients A acts only on index a,
        # this index is then summed, and then each mode of r, A@r undergoes
        # matrix multiplication
        ind1 = indices[: 2 * num_modes + 1]
        ind2 = ind1[0] + "".join(
            [
                str(i) + str(j)
                for i, j in zip(ind1[2::2], indices[2 * num_modes + 1 : 3 * num_modes + 1])
            ]
        )
        ind3 = "".join([str(i) + str(j) for i, j in zip(ind1[1::2], ind2[2::2])])
        ind = "{},{}->{}".format(ind1, ind2, ind3)

        if np.allclose(quad_coeffs, 0.0):
            poly_op = np.zeros([dim] * (2 * num_modes), dtype=np.complex128)
        else:
            # Einsum above applied to to r,Ar
            # This einsum sums over all quadrature operators, and also applies matrix
            # multiplication between the same mode of each operator
            poly_op = np.einsum(ind, r, np.tensordot(quad_coeffs, r, axes=1)).conj()

        # add linear term
        rows = np.flip(np.array(rows).reshape([2, -1]), axis=1).flatten()
        poly_op += r.T @ linear_coeff[rows]

        # add constant term
        if k != 0:
            poly_op += k * expand_dims(np.eye(dim), 0, num_modes)

        # calculate Op^2
        ind = "{},{}->{}".format(ind1[1:], ind2[1:], ind3)
        poly_op_sq = np.einsum(ind, poly_op, poly_op)

        # truncate down
        sl = tuple([slice(0, self._cutoff)] * (2 * num_modes))
        poly_op = poly_op[sl]
        poly_op_sq = poly_op_sq[sl]

        ind1 = ind1[:-1]
        ind2 = "".join([str(j) + str(i) for i, j in zip(ind1[::2], ind1[1::2])])
        ind = "{},{}".format(ind1, ind2)

        # calculate expectation value, Tr(Op @ rho)
        # For 3 modes, this gives the einsum '(ab)(cd)(ef),(ba)(dc)(fe)->'
        mean = np.einsum(ind, poly_op, rho).real

        # calculate variance Tr(Op^2 @ rho) - Tr(Op @ rho)^2
        var = np.einsum(ind, poly_op_sq, rho).real - mean ** 2

        return mean, var


class BaseGaussianState(BaseState):
    r"""Class for the representation of quantum states using the Gaussian formalism.

    Note that this class uses the Gaussian representation convention

    .. math:: \bar{\mathbf{r}} = (\bar{x}_1,\bar{x}_2,\dots,\bar{x}_N,\bar{p}_1,\dots,\bar{p}_N)

    Args:
        state_data (tuple(mu, cov)): A tuple containing the vector of means array ``mu`` and the
            covariance matrix array ``cov``, in terms of the complex displacement.
        num_modes (int): the number of modes in the state
        pure (bool): True if the state is a pure state, false if the state is mixed
        mode_names (Sequence): (optional) this argument contains a list providing mode names
            for each mode in the state
    """

    def __init__(self, state_data, num_modes, mode_names=None):
        super().__init__(num_modes, mode_names)

        self._data = state_data

        # vector of means and covariance matrix, using frontend x,p scaling
        self._mu = self._data[0] * np.sqrt(self._hbar / 2)
        self._cov = self._data[1] * (self._hbar / 2)
        # complex displacements of the Gaussian state
        self._alpha = self._mu[: self._modes] + 1j * self._mu[self._modes :]
        self._alpha /= np.sqrt(2 * self._hbar)

        self._pure = (
            np.abs(np.linalg.det(self._cov) - (self._hbar / 2) ** (2 * self._modes))
            < self.EQ_TOLERANCE
        )

        self._basis = "gaussian"
        self._str = "<GaussianState: num_modes={}, pure={}, hbar={}>".format(
            self.num_modes, self._pure, self._hbar
        )

    def __eq__(self, other):
        """Equality operator for BaseGaussianState.

        Returns True if other BaseGaussianState is close to self.
        This is done by comparing the means vector and cov matrix.
        If both are within the EQ_TOLERANCE, True is returned.

        Args:
            other (BaseGaussianState): BaseGaussianState to compare against.
        """
        # pylint: disable=protected-access
        if not isinstance(other, type(self)):
            return False

        if self.num_modes != other.num_modes:
            return False

        if np.allclose(self._mu, other._mu, atol=self.EQ_TOLERANCE, rtol=0) and np.allclose(
            self._cov, other._cov, atol=self.EQ_TOLERANCE, rtol=0
        ):
            return True

        return False

    def means(self):
        r"""The vector of means describing the Gaussian state.

        For a :math:`N` mode state, this has the form

        .. math::
            \bar{\mathbf{r}} = \left(\bar{x}_0,\dots,\bar{x}_{N-1},\bar{p}_0,\dots,\bar{p}_{N-1}\right)

        where :math:`\bar{x}_i` and :math:`\bar{p}_i` refer to the mean
        position and momentum quadrature of mode :math:`i` respectively.

        Returns:
          array: a length :math:`2N` array containing the vector of means.
        """
        return self._mu

    def cov(self):
        r"""The covariance matrix describing the Gaussian state.

        The diagonal elements of the covariance matrix correspond to the
        variance in the position and momentum quadratures:

        .. math::
            \mathbf{V}_{ii} = \begin{cases}
                (\Delta x_i)^2, & 0\leq i\leq N-1\\
                (\Delta p_{i-N})^2, & N\leq i\leq 2(N-1)
            \end{cases}

        where :math:`\Delta x_i` and :math:`\Delta p_i` refer to the
        position and momentum quadrature variance of mode :math:`i` respectively.

        Note that if the covariance matrix is purely diagonal, then this
        corresponds to squeezing :math:`z=re^{i\phi}` where :math:`\phi=0`,
        and :math:`\Delta x_i = e^{-2r}`, :math:`\Delta p_i = e^{2r}`.

        Returns:
          array: the :math:`2N\times 2N` covariance matrix.
        """
        return self._cov

    def reduced_gaussian(self, modes):
        r""" Returns the vector of means and the covariance matrix of the specified modes.

        Args:
            modes (int of Sequence[int]): indices of the requested modes

        Returns:
            tuple (means, cov): where means is an array containing the vector of means,
            and cov is a square array containing the covariance matrix.
        """
        if modes == list(range(self._modes)):
            # reduced state is full state
            return self._mu, self._cov

        # reduce rho down to specified subsystems
        if isinstance(modes, int):
            modes = [modes]

        if modes != sorted(modes):
            raise ValueError("The specified modes cannot be duplicated.")

        if len(modes) > self._modes:
            raise ValueError(
                "The number of specified modes cannot " "be larger than the number of subsystems."
            )

        ind = np.concatenate([np.array(modes), np.array(modes) + self._modes])
        rows = ind.reshape(-1, 1)
        cols = ind.reshape(1, -1)

        mu = self._mu[ind]
        cov = self._cov[rows, cols]

        return mu, cov

    def is_coherent(self, mode, tol=1e-10):
        r"""Returns True if the Gaussian state of a particular mode is a coherent state.

        Args:
            mode (int): the specified mode
            tol (float): the numerical precision in determining if squeezing is not present

        Returns:
            bool: True if and only if the state is a coherent state.
        """
        mu, cov = self.reduced_gaussian([mode])  # pylint: disable=unused-variable
        cov /= self._hbar / 2
        return np.allclose(cov, np.identity(2), atol=tol, rtol=0)

    def displacement(self, modes=None):
        r"""Returns the displacement parameter :math:`\alpha` of the modes specified.

        Args:
            modes (int or Sequence[int]): modes specified

        Returns:
            Sequence[complex]: sequence of complex displacements :math:`\alpha`
            corresponding to the list of specified modes
        """
        if modes is None:
            modes = list(range(self._modes))
        elif isinstance(modes, int):  # pragma: no cover
            modes = [modes]

        return self._alpha[list(modes)]

    def is_squeezed(self, mode, tol=1e-6):
        r"""Returns True if the Gaussian state of a particular mode is a squeezed state.

        Args:
            mode (int): the specified mode
            tol (float): the numerical precision in determining if squeezing is present

        Returns:
           bool: True if and only if the state is a squeezed state.
        """
        mu, cov = self.reduced_gaussian([mode])  # pylint: disable=unused-variable
        cov /= self._hbar / 2
        return np.any(np.abs(cov - np.identity(2)) > tol)

    def squeezing(self, modes=None):
        r"""Returns the squeezing parameters :math:`(r,\phi)` of the modes specified.

        Args:
            modes (int or Sequence[int]): modes specified

        Returns:
            List[(float, float)]: sequence of tuples containing the squeezing
            parameters :math:`(r,\phi)` of the specified modes.
        """
        if modes is None:
            modes = list(range(self._modes))
        elif isinstance(modes, int):  # pragma: no cover
            modes = [modes]

        res = []
        for i in modes:
            mu, cov = self.reduced_gaussian([i])  # pylint: disable=unused-variable
            cov /= self._hbar / 2
            tr = np.trace(cov)

            r = np.arccosh(tr / 2) / 2

            if cov[0, 1] == 0.0:
                phi = 0
            else:
                phi = -np.arcsin(2 * cov[0, 1] / np.sqrt((tr - 2) * (tr + 2)))

            res.append((r, phi))

        return res

    # =====================================================
    # the following methods are overwritten from BaseState

    def wigner(self, mode, xvec, pvec):
        mu, cov = self.reduced_gaussian([mode])

        X, P = np.meshgrid(xvec, pvec)
        grid = np.empty(X.shape + (2,))
        grid[:, :, 0] = X
        grid[:, :, 1] = P
        mvn = multivariate_normal(mu, cov, allow_singular=True)

        return mvn.pdf(grid)

    def quad_expectation(self, mode, phi=0, **kwargs):
        # pylint: disable=unused-argument
        mu, cov = self.reduced_gaussian([mode])
        rot = _R(phi)

        muphi = rot.T @ mu
        covphi = rot.T @ cov @ rot
        return (muphi[0], covphi[0, 0])

    def poly_quad_expectation(self, A, d=None, k=0, phi=0, **kwargs):
        if A is None:
            A = np.zeros([2 * self._modes, 2 * self._modes])

        if A.shape != (2 * self._modes, 2 * self._modes):
            raise ValueError("Matrix of quadratic coefficients A must be of size 2Nx2N.")

        if not np.allclose(A.T, A):
            raise ValueError("Matrix of quadratic coefficients A must be symmetric.")

        if d is not None:
            if d.shape != (2 * self._modes,):
                raise ValueError("Vector of linear coefficients d must be of length 2N.")
        else:
            d = np.zeros([2 * self._modes])

        # determine modes with quadratic expectation values
        nonzero = np.concatenate(
            [np.mod(A.nonzero()[0], self._modes), np.mod(d.nonzero()[0], self._modes)]
        )
        ex_modes = list(set(nonzero))

        # reduce the size of A so that we only consider modes
        # which we need to calculate the expectation value for
        rows = ex_modes + [i + self._modes for i in ex_modes]
        num_modes = len(ex_modes)
        quad_coeffs = A[:, rows][rows]

        if not ex_modes:
            # only a constant term was provided
            return k, 0.0

        mu = self._mu
        cov = self._cov

        if phi != 0:
            # rotate all modes of the covariance matrix and vector of means
            R = _R(phi)
            C = changebasis(self._modes)
            rot = C.T @ block_diag(*([R] * self._modes)) @ C

            mu = rot.T @ mu
            cov = rot.T @ cov @ rot

        # transform to the expectation of a quadratic on a normal distribution with zero mean
        # E[P(r)]_(mu,cov) = E(Q(r+mu)]_(0,cov)
        #                  = E[rT.A.r + rT.(2A.mu+d) + (muT.A.mu+muT.d+cI)]_(0,cov)
        #                  = E[rT.A.r + rT.d' + k']_(0,cov)
        d2 = 2 * A @ mu + d
        k2 = mu.T @ A @ mu + mu.T @ d + k

        # expectation value E[P(r)]_{mu=0} = tr(A.cov) + muT.A.mu + muT.d + k|_{mu=0}
        #                                  = tr(A.cov) + k
        mean = np.trace(A @ cov) + k2
        # variance Var[P(r)]_{mu=0} = 2tr(A.cov.A.cov) + 4*muT.A.cov.A.mu + dT.cov.d|_{mu=0}
        #                           = 2tr(A.cov.A.cov) + dT.cov.d
        var = 2 * np.trace(A @ cov @ A @ cov) + d2.T @ cov @ d2

        # Correction term to account for incorrect symmetric ordering in the variance.
        # This occurs because Var[S(P(r))] = Var[P(r)] - Σ_{m1, m2} |hbar*A_{(m1, m1+N),(m2, m2+N)}|,
        # where m1, m2 are all possible mode numbers, and N is the total number of modes.
        # Therefore, the correction term is the sum of the determinants of 2x2 submatrices of A.
        modes = np.arange(2 * num_modes).reshape(2, -1).T
        var -= np.sum(
            [np.linalg.det(self._hbar * quad_coeffs[:, m][n]) for m in modes for n in modes]
        )

        return mean, var

    @abc.abstractmethod
    def reduced_dm(self, modes, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def fock_prob(self, n, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def mean_photon(self, mode, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def fidelity(self, other_state, mode, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def fidelity_vacuum(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def fidelity_coherent(self, alpha_list, **kwargs):
        raise NotImplementedError



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Tools for users to identify large cliques in graphs.

A clique is a subgraph where all nodes are connected to each other. The maximum clique problem is
to identify the largest clique in a graph. It has been shown that samples from GBS can be used to
select dense subgraphs as a starting seed for heuristic algorithms :cite:`banchi2019molecular`.

.. seealso::

    :ref:`apps-clique-tutorial`

Algorithm
^^^^^^^^^

This module provides :func:`search`, a variant of the local search heuristics described in
:cite:`pullan2006dynamic` and :cite:`pullan2006phased`. The algorithm proceeds as follows:

#. A small clique in the graph is identified. The initial clique can be a single node, or it can
   be obtained by shrinking a random subgraph, for example obtained from GBS.
#. The clique is grown as much as possible by adding nodes connected to all nodes in the clique.
#. When no more growth is possible, a node in the clique is swapped with a node outside of it.

Steps 2-3 are repeated until a pre-determined number of steps have been completed, or until no
more growth or swaps are possible.

Clique growth and swapping
^^^^^^^^^^^^^^^^^^^^^^^^^^

A clique can be grown by evaluating the set :math:`C_0` of nodes in the remainder of the graph that
are connected to all nodes in the clique. A single node from :math:`C_0` is selected and added to
the clique. This process is repeated until :math:`C_0` becomes empty. This is provided with the
:func:`grow` function.

Searching the local space around a clique can be achieved by swapping a node from the clique with a
node in the remainder of the graph. The first step is to evaluate the set :math:`C_1` of nodes in
the remainder of the graph that are connected to *all but one* of the nodes in the current
clique. A swap is then performed by adding the node into the clique and removing the node in the
clique that is not connected to it. This is provided with the :func:`swap` function.

Whenever the sets :math:`C_0` and :math:`C_1` used during growth and swapping have more than
one element, there must be a choice of which node to add or swap. The supported choices are:

- Select among candidate nodes uniformly at random;
- Select the candidate node with the greatest degree, settling remaining ties uniformly at random;
- Select the candidate node with the greatest node weight, settling remaining ties uniformly at
  random.

Using GBS to find a starting clique
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Samples from GBS correspond to subgraphs that are likely to be dense.
These subgraphs may not be cliques, which is the required input to the :func:`search` algorithm.
To reconcile this, a subgraph may be shrunk using the :func:`shrink` function by removing nodes
until the remainder forms a clique. This can be achieved by selecting the node with the lowest
degree relative to the rest of subgraph and removing the node, repeating the process until a
clique is found.
"""
from typing import Union

import networkx as nx
import numpy as np


def search(
    clique: list, graph: nx.Graph, iterations, node_select: Union[str, np.ndarray, list] = "uniform"
) -> list:
    """Local search algorithm for identifying large cliques.

    This function implements a version of the local search algorithm given in
    :cite:`pullan2006dynamic` and :cite:`pullan2006phased`. It proceeds by iteratively applying
    phases of greedy growth and plateau search to the input clique subgraph.

    **Growth phase**

    Growth is achieved using the :func:`~.grow` function, which repeatedly evaluates the
    set :math:`C_0` of nodes in the remainder of the graph that are connected to all nodes in the
    clique, and selects one candidate node from :math:`C_0` to make a larger clique.

    **Search phase**

    Plateau search is performed with the :func:`~.swap` function, which evaluates the set
    :math:`C_1` of nodes in the remainder of the graph that are connected to all but one of the
    nodes in the clique. The function then proceeds to select one candidate from :math:`C_1` and
    swap it with its corresponding node in the clique.

    Whenever the sets :math:`C_0` and :math:`C_1` used during growth and swapping have more than
    one element, there must be a choice of which node to add or swap. This choice is specified
    with the ``node_select`` argument, which can be any of the following:

    - ``"uniform"`` (default): choose a node from the candidates uniformly at random;
    - ``"degree"``: choose the node from the candidates with the greatest degree, settling ties
      by uniform random choice;
    - A list or array: specifying the node weights of the graph, resulting in choosing the node
      from the candidates with the greatest weight, settling ties by uniform random choice.

    **Stopping**

    The iterative applications of growth and search are applied until either the specified number
    of iterations has been carried out or when a dead end has been reached. The function reaches
    a dead end when it is unable to perform any new swaps or growth.

    **Example usage:**

    >>> graph = nx.lollipop_graph(4, 1)
    >>> graph.add_edge(2, 4)
    >>> clique = [3, 4]
    >>> search(clique, graph, iterations=10)
    [0, 1, 2, 3]

    Args:
        clique (list[int]): a subgraph specified by a list of nodes; the subgraph must be a clique
        graph (nx.Graph): the input graph
        iterations (int): number of steps in the algorithm
        node_select (str, list or array): method of selecting nodes during swap and growth. Can
            be ``"uniform"`` (default), ``"degree"``, or a NumPy array or list.

    Returns:
       list[int]: the largest clique found by the algorithm
    """
    if iterations < 1:
        raise ValueError("Number of iterations must be a positive int")

    grown = grow(clique, graph, node_select=node_select)
    swapped = swap(grown, graph, node_select=node_select)

    iterations -= 1

    if set(grown) == set(swapped) or iterations == 0:
        return swapped

    return search(swapped, graph, iterations, node_select)


def grow(
    clique: list, graph: nx.Graph, node_select: Union[str, np.ndarray, list] = "uniform"
) -> list:
    """Iteratively adds new nodes to the input clique to generate a larger clique.

    Each iteration involves calculating the set :math:`C_0` (provided by the function
    :func:`c_0`) with respect to the current clique. This set represents the nodes in the rest of
    the graph that are connected to all of the nodes in the current clique. Therefore, adding any of
    the nodes in :math:`C_0` will create a larger clique. This function proceeds by repeatedly
    evaluating :math:`C_0` and selecting and adding a node from this set to add to the current
    clique. Growth is continued until :math:`C_0` becomes empty.

    Whenever there are multiple nodes within :math:`C_0`, one must choose which node to add to
    the growing clique. This function allows a method of choosing nodes to be set with the
    ``node_select`` argument, which can be any of the following:

    - ``"uniform"`` (default): choose a node from the candidates uniformly at random;
    - ``"degree"``: choose the node from the candidates with the greatest degree, settling ties
      by uniform random choice;
    - A list or array: specifying the node weights of the graph, resulting in choosing the node
      from the candidates with the greatest weight, settling ties by uniform random choice.

    **Example usage:**

    >>> graph = nx.complete_graph(10)
    >>> clique = [0, 1, 2, 3, 4]
    >>> grow(clique, graph)
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    Args:
        clique (list[int]): a subgraph specified by a list of nodes; the subgraph must be a clique
        graph (nx.Graph): the input graph
        node_select (str, list or array): method of selecting nodes from :math:`C_0` during
            growth. Can be ``"uniform"`` (default), ``"degree"``, or a NumPy array or list.

    Returns:
        list[int]: a new clique subgraph of equal or larger size than the input
    """

    if not set(clique).issubset(graph.nodes):
        raise ValueError("Input is not a valid subgraph")

    if not is_clique(graph.subgraph(clique)):
        raise ValueError("Input subgraph is not a clique")

    if isinstance(node_select, (list, np.ndarray)):
        if len(node_select) != graph.number_of_nodes():
            raise ValueError("Number of node weights must match number of nodes")
        w = {n: node_select[i] for i, n in enumerate(graph.nodes)}
        node_select = "weight"

    clique = set(clique)
    _c_0 = sorted(c_0(clique, graph))

    while _c_0:
        if node_select == "uniform":
            clique.add(np.random.choice(_c_0))
        elif node_select == "degree":
            degrees = np.array([graph.degree(n) for n in _c_0])
            to_add_index = np.random.choice(np.where(degrees == degrees.max())[0])
            to_add = _c_0[to_add_index]
            clique.add(to_add)
        elif node_select == "weight":
            weights = np.array([w[n] for n in _c_0])
            to_add_index = np.random.choice(np.where(weights == weights.max())[0])
            to_add = _c_0[to_add_index]
            clique.add(to_add)
        else:
            raise ValueError("Node selection method not recognized")

        _c_0 = sorted(c_0(clique, graph))

    return sorted(clique)


def swap(
    clique: list, graph: nx.Graph, node_select: Union[str, np.ndarray, list] = "uniform"
) -> list:
    """If possible, generates a new clique by swapping a node in the input clique with a node
    outside the clique.

    Proceeds by calculating the set :math:`C_1` of nodes in the rest of the graph that are
    connected to all but one of the nodes in the clique. If this set is not empty, this function
    randomly picks a node and swaps it with the corresponding node in the clique that is not
    connected to it. The set :math:`C_1` and corresponding nodes in the clique are provided by the
    :func:`c_1` function.

    Whenever there are multiple nodes within :math:`C_1`, one must choose which node to add to
    the growing clique. This function allows a method of choosing nodes to be set with the
    ``node_select`` argument, which can be any of the following:

    - ``"uniform"`` (default): choose a node from the candidates uniformly at random;
    - ``"degree"``: choose the node from the candidates with the greatest degree, settling ties
      by uniform random choice;
    - A list or array: specifying the node weights of the graph, resulting in choosing the node
      from the candidates with the greatest weight, settling ties by uniform random choice.

    **Example usage:**

    >>> graph = nx.wheel_graph(5)
    >>> graph.remove_edge(0, 4)
    >>> clique = [0, 1, 2]
    >>> swap(clique, graph)
    [0, 2, 3]

    Args:
        clique (list[int]): a subgraph specified by a list of nodes; the subgraph must be a clique
        graph (nx.Graph): the input graph
        node_select (str, list or array): method of selecting incoming nodes from :math:`C_1`
            during swapping. Can be ``"uniform"`` (default), ``"degree"``, or a NumPy array or list.

    Returns:
        list[int]: a new clique subgraph of equal size as the input
    """

    if not set(clique).issubset(graph.nodes):
        raise ValueError("Input is not a valid subgraph")

    if not is_clique(graph.subgraph(clique)):
        raise ValueError("Input subgraph is not a clique")

    if isinstance(node_select, (list, np.ndarray)):
        if len(node_select) != graph.number_of_nodes():
            raise ValueError("Number of node weights must match number of nodes")
        w = {n: node_select[i] for i, n in enumerate(graph.nodes)}
        node_select = "weight"

    clique = set(clique)
    _c_1 = c_1(clique, graph)

    if _c_1:
        if node_select == "uniform":
            swap_index = np.random.choice(len(_c_1))
            swap_nodes = _c_1[swap_index]
        elif node_select == "degree":
            degrees = np.array([graph.degree(n[1]) for n in _c_1])
            to_swap_index = np.random.choice(np.where(degrees == degrees.max())[0])
            swap_nodes = _c_1[to_swap_index]
        elif node_select == "weight":
            weights = np.array([w[n[1]] for n in _c_1])
            to_swap_index = np.random.choice(np.where(weights == weights.max())[0])
            swap_nodes = _c_1[to_swap_index]
        else:
            raise ValueError("Node selection method not recognized")

        clique.remove(swap_nodes[0])
        clique.add(swap_nodes[1])

    return sorted(clique)


def shrink(
    subgraph: list, graph: nx.Graph, node_select: Union[str, np.ndarray, list] = "uniform"
) -> list:
    """Shrinks an input subgraph until it forms a clique.

    Proceeds by removing nodes in the input subgraph one at a time until the result is a clique
    that satisfies :func:`is_clique`. Upon each iteration, this function selects the node with
    lowest degree relative to the subgraph and removes it.

    In some instances, there may be multiple nodes of minimum degree as candidates to remove from
    the subgraph. The method of selecting which of these nodes to remove is specified by the
    ``node_select`` argument, which can be either:

    - ``"uniform"`` (default): choose a node from the candidates uniformly at random;
    - A list or array: specifying the node weights of the graph, resulting in choosing the node
      from the candidates with the lowest weight, settling ties by uniform random choice.

    **Example usage:**

    >>> graph = nx.barbell_graph(4, 0)
    >>> subgraph = [0, 1, 2, 3, 4, 5]
    >>> shrink(subgraph, graph)
    [0, 1, 2, 3]

    Args:
        subgraph (list[int]): a subgraph specified by a list of nodes
        graph (nx.Graph): the input graph
        node_select (str, list or array): method of settling ties when more than one node of
            equal degree can be removed. Can be ``"uniform"`` (default), or a NumPy array or list.

    Returns:
        list[int]: a clique of size smaller than or equal to the input subgraph
    """

    if not set(subgraph).issubset(graph.nodes):
        raise ValueError("Input is not a valid subgraph")

    if isinstance(node_select, (list, np.ndarray)):
        if len(node_select) != graph.number_of_nodes():
            raise ValueError("Number of node weights must match number of nodes")
        w = {n: node_select[i] for i, n in enumerate(graph.nodes)}
        node_select = "weight"

    subgraph = graph.subgraph(subgraph).copy()  # A copy is required to be able to modify the
    # structure of the subgraph (https://networkx.github.io/documentation/stable/reference/classes/generated/networkx.Graph.subgraph.html)

    while not is_clique(subgraph):
        degrees = np.array(subgraph.degree)
        degrees_min = np.argwhere(degrees[:, 1] == degrees[:, 1].min()).flatten()

        if node_select == "uniform":
            to_remove_index = np.random.choice(degrees_min)
        elif node_select == "weight":
            weights = np.array([w[degrees[n][0]] for n in degrees_min])
            to_remove_index = np.random.choice(np.where(weights == weights.min())[0])
        else:
            raise ValueError("Node selection method not recognized")

        to_remove = degrees[to_remove_index][0]
        subgraph.remove_node(to_remove)

    return sorted(subgraph.nodes())


def is_clique(graph: nx.Graph) -> bool:
    """Determines if the input graph is a clique. A clique of :math:`n` nodes has exactly :math:`n(
    n-1)/2` edges.

    **Example usage:**

    >>> graph = nx.complete_graph(10)
    >>> is_clique(graph)
    True

    Args:
        graph (nx.Graph): the input graph

    Returns:
        bool: ``True`` if input graph is a clique and ``False`` otherwise
    """
    edges = graph.edges
    nodes = graph.order()

    return len(edges) == nodes * (nodes - 1) / 2


def c_0(clique: list, graph: nx.Graph):
    """Generates the set :math:`C_0` of nodes that are connected to all nodes in the input
    clique subgraph.

    The set :math:`C_0` is defined in :cite:`pullan2006phased` and is used to determine nodes
    that can be added to the current clique to grow it into a larger one.

    **Example usage:**

    >>> graph = nx.complete_graph(10)
    >>> clique = [0, 1, 2, 3, 4]
    >>> c_0(clique, graph)
    [5, 6, 7, 8, 9]

    Args:
        clique (list[int]): a subgraph specified by a list of nodes; the subgraph must be a clique
        graph (nx.Graph): the input graph

    Returns:
        list[int]: a list containing the :math:`C_0` nodes for the clique

    """
    if not is_clique(graph.subgraph(clique)):
        raise ValueError("Input subgraph is not a clique")

    clique = set(clique)
    c_0_nodes = []
    non_clique_nodes = set(graph.nodes) - clique

    for i in non_clique_nodes:
        if clique.issubset(graph.neighbors(i)):
            c_0_nodes.append(i)

    return c_0_nodes


def c_1(clique: list, graph: nx.Graph):
    """Generates the set :math:`C_1` of nodes that are connected to all but one of the nodes in
    the input clique subgraph.

    The set :math:`C_1` is defined in :cite:`pullan2006phased` and is used to determine outside
    nodes that can be swapped with clique nodes to create a new clique.

    **Example usage:**

    >>> graph = nx.wheel_graph(5)
    >>> clique = [0, 1, 2]
    >>> c_1(clique, graph)
    [(1, 3), (2, 4)]

    Args:
        clique (list[int]): a subgraph specified by a list of nodes; the subgraph must be a clique
        graph (nx.Graph): the input graph

    Returns:
       list[tuple[int]]: A list of tuples. The first node in the tuple is the node in the clique
       and the second node is the outside node it can be swapped with.
   """
    if not is_clique(graph.subgraph(clique)):
        raise ValueError("Input subgraph is not a clique")

    clique = set(clique)
    c_1_nodes = []
    non_clique_nodes = set(graph.nodes) - clique

    for i in non_clique_nodes:
        neighbors_in_subgraph = clique.intersection(graph.neighbors(i))

        if len(neighbors_in_subgraph) == len(clique) - 1:
            to_swap = clique - neighbors_in_subgraph
            (i_clique,) = to_swap
            c_1_nodes.append((i_clique, i))

    return c_1_nodes



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Circuit specifications for the Gaussian simulator backend."""
from .circuit_specs import CircuitSpecs


class GaussianSpecs(CircuitSpecs):
    """Circuit specifications for the Gaussian backend."""

    short_name = "gaussian"
    modes = None
    local = True
    remote = True
    interactive = True

    primitives = {
        # meta operations
        "All",
        "_New_modes",
        "_Delete",
        # state preparations
        "Vacuum",
        "Coherent",
        "Squeezed",
        "DisplacedSqueezed",
        "Thermal",
        "Gaussian",
        # measurements
        "MeasureHomodyne",
        "MeasureHeterodyne",
        "MeasureFock",
        "MeasureThreshold",
        # channels
        "LossChannel",
        "ThermalLossChannel",
        # single mode gates
        "Dgate",
        "Sgate",
        "Rgate",
        "BSgate",
    }

    decompositions = {
        "Interferometer": {},
        "GraphEmbed": {},
        "BipartiteGraphEmbed": {},
        "GaussianTransform": {},
        "Gaussian": {},
        "Pgate": {},
        "S2gate": {},
        "CXgate": {},
        "CZgate": {},
        "MZgate": {},
        "Xgate": {},
        "Zgate": {},
        "Fouriergate": {},
    }



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tensorflow backend proper
======================

Contains most of the code for managing the simulator state and offloading
operations to the utilities in ops.

Hyperlinks: :class:`Circuit`

.. currentmodule:: strawberryfields.backends.tfbackend.circuit

Contents
----------------------
.. autosummary::
     Circuit

"""
# pylint: disable=too-many-arguments,too-many-statements,too-many-branches,protected-access,attribute-defined-outside-init

import numbers
from itertools import product
from string import ascii_lowercase as indices

import numpy as np
from scipy.special import factorial
import tensorflow as tf

from . import ops


class Circuit:
    """Base class for representing and operating on a collection of
         CV quantum optics modes in the Fock basis.
         The modes are initialized in the (multimode) vacuum state,
         using the Fock representation with given cutoff_dim.
         The state of the modes is manipulated by calling the various methods."""

    # pylint: disable=too-many-instance-attributes,too-many-public-methods
    def __init__(self, graph, num_modes, cutoff_dim, pure, batch_size):
        self._graph = (
            None  # will be set when reset is called below, but reset needs something to compare to
        )
        self._hbar = 2
        self.reset(
            graph=graph,
            num_subsystems=num_modes,
            pure=pure,
            cutoff_dim=cutoff_dim,
            batch_size=batch_size,
        )

    def _make_vac_states(self, cutoff_dim):
        """Make vacuum state tensors for the underlying graph"""
        with self._graph.as_default():
            one = tf.cast([1.0], ops.def_type)
            v = tf.scatter_nd([[0]], one, [cutoff_dim])
            self._single_mode_pure_vac = v
            self._single_mode_mixed_vac = tf.einsum("i,j->ij", v, v)
            if self._batched:
                self._single_mode_pure_vac = tf.stack(
                    [self._single_mode_pure_vac] * self._batch_size
                )
                self._single_mode_mixed_vac = tf.stack(
                    [self._single_mode_mixed_vac] * self._batch_size
                )

    def _update_state(self, new_state):
        """Helper function to update the state history and the current state"""
        # pylint: disable=attribute-defined-outside-init
        self._state_history.append(new_state)
        self._state = new_state

    def _valid_modes(self, modes):
        # todo: this method should probably be moved into BaseBackend and then maybe
        # overridden and expended in the subclasses to avoid code duplication and
        # missing out on conditions.
        if isinstance(modes, int):
            modes = [modes]

        for mode in modes:
            if mode < 0:
                raise ValueError("Specified mode number(s) cannot be negative.")
            elif mode >= self._num_modes:
                raise ValueError(
                    "Specified mode number(s) are not compatible with number of modes."
                )

        if len(modes) != len(set(modes)):
            raise ValueError("The specified modes cannot appear multiple times.")

        return True

    def _replace_and_update(self, replacement, modes):
        """
        Helper function for replacing a mode, updating the state history,
        and possibly setting circuit's state_is_pure variable to a new value.

        Expects replacement to be batched if self._batched.
        """
        if isinstance(modes, int):
            modes = [modes]

        if self._batched:
            batch_offset = 1
        else:
            batch_offset = 0

        num_modes = len(self._state.shape) - batch_offset
        if not self._state_is_pure:
            num_modes = int(num_modes / 2)

        new_state = ops.replace_modes(
            replacement, modes, self._state, self._state_is_pure, self._batched
        )
        self._update_state(new_state)

        # update purity depending on whether we have replaced all modes or a subset
        if len(modes) == num_modes:
            replacement_is_pure = bool(len(replacement.shape) - batch_offset == len(modes))
            self._state_is_pure = replacement_is_pure
        else:
            self._state_is_pure = False

    def _maybe_batch(self, param, convert_to_tensor=True):
        """Helper function to broadcast a param to the correct shape (if necessary) when working in batch mode. If param is not a scalar,
        it will raise an exception if param's batch size is not equal to the circuit's batch size."""
        shape_err = False
        if convert_to_tensor:
            p = tf.convert_to_tensor(param)
            if not self._batched:
                broadcast_p = p
            else:
                if p.shape.ndims == 0:
                    # scalar
                    broadcast_p = tf.stack([p] * self._batch_size)
                elif p.shape.ndims == 1:
                    # vector
                    if p.shape.dims[0].value == self._batch_size:
                        broadcast_p = p
                    else:
                        shape_err = True
                else:
                    shape_err = True
        else:
            p = np.array(param)
            if not self._batched:
                broadcast_p = p
            else:
                if len(p.shape) == 0:  # pylint: disable=len-as-condition
                    # scalar
                    broadcast_p = np.concatenate([np.expand_dims(p, 0)] * self._batch_size)
                elif len(p.shape) == 1:
                    # vector
                    if p.shape[0] == self._batch_size:
                        broadcast_p = p
                    else:
                        shape_err = True
                else:
                    shape_err = True

        if shape_err:
            raise ValueError(
                "Parameter can be either a scalar or a vector of length {}.".format(
                    self._batch_size
                )
            )
        else:
            return broadcast_p

    def _check_incompatible_batches(self, *params):
        """Helper function for verifying that all the params from a list have the same batch size. Only does something
             when the circuit is running in batched mode."""
        if self._batched:
            for idx, p in enumerate(params):
                param_batch_size = p.shape.dims[0].value
                if idx == 0:
                    ref_batch_size = param_batch_size
                else:
                    if param_batch_size != ref_batch_size:
                        raise ValueError("Parameters have incompatible batch sizes.")

    def del_mode(self, modes_list):
        """Remove the modes in modes_list from the circuit."""
        pure = self._state_is_pure
        for mode in modes_list:
            reduced_state = ops.partial_trace(self._state, mode, pure, self._batched)
            pure = False
        self._update_state(reduced_state)
        self._state_is_pure = False
        self._num_modes -= len(modes_list)

    def add_mode(self, num_modes):
        """Append M modes (initialized in vacuum states) to the circuit."""
        vac = self._single_mode_pure_vac if self._state_is_pure else self._single_mode_mixed_vac
        new_state = self._state
        for _ in range(num_modes):
            new_state = ops.insert_state(vac, new_state, self._state_is_pure, batched=self._batched)
        self._update_state(new_state)
        self._num_modes += num_modes

    def reset(self, **kwargs):
        r"""
        Resets the state of the circuit to have all modes in vacuum.
        If a keyword arg is not present, the corresponding parameter is unchanged.

        Keyword Args:
            graph (tf.Graph): the underlying graph (and any associated attributes) is replaced with this supplied graph. If None, the same underlying
              graph (and all its defined operations) will be kept.
            num_subsystems (int): sets the number of modes in the reset circuit.
            pure (bool): if True, the reset circuit will represent its state as a pure state. If False, the representation will be mixed.
            cutoff_dim (int): new Fock space cutoff dimension to use.
            batch_size (None, int): None means no batching. An integer value >= 2 sets the batch size to use.
        """
        if "pure" in kwargs:
            pure = kwargs["pure"]
            if not isinstance(pure, bool):
                raise ValueError("Argument 'pure' must be either True or False")
            self._state_is_pure = pure

        if "num_subsystems" in kwargs:
            ns = kwargs["num_subsystems"]
            if not isinstance(ns, int):
                raise ValueError("Argument 'num_subsystems' must be a positive integer")
            self._num_modes = ns

        if "cutoff_dim" in kwargs:
            cutoff_dim = kwargs["cutoff_dim"]
            if not isinstance(cutoff_dim, int) or cutoff_dim < 1:
                raise ValueError("Argument 'cutoff_dim' must be a positive integer")
            self._cutoff_dim = cutoff_dim

        if "batch_size" in kwargs:
            batch_size = kwargs["batch_size"]
            if batch_size is not None:
                if not isinstance(batch_size, int) or batch_size < 2:
                    raise ValueError("Argument 'batch_size' must be either None or an integer > 1")
            self._batch_size = batch_size
            self._batched = batch_size is not None

        if "graph" in kwargs:
            graph = kwargs["graph"]
            if not isinstance(graph, tf.Graph):
                raise ValueError("Argument 'graph' must be a tf.Graph")
            if graph != self._graph:
                self._graph = graph
                ops.get_prefac_tensor.cache_clear()  # clear any cached tensors that may live on old graph
            self._state_history = []
            self._cache = {}

        with self._graph.as_default():
            # todo vac states only need to be remade if graph or cutoff_dim changes?
            self._make_vac_states(self._cutoff_dim)
            single_mode_vac = self._single_mode_pure_vac if pure else self._single_mode_mixed_vac
            if self._num_modes == 1:
                vac = single_mode_vac
            else:
                vac = ops.combine_single_modes(
                    [single_mode_vac] * self._num_modes, self._batch_size
                )
            vac = tf.identity(vac, name="Vacuum")
            self._update_state(vac)

    def prepare_vacuum_state(self, mode):
        """
        Traces out the state in 'mode' and replaces it with a vacuum state.
        """
        if self._valid_modes(mode):
            with self._graph.as_default():
                if self._state_is_pure:
                    state = self._single_mode_pure_vac
                else:
                    state = self._single_mode_mixed_vac
                self._replace_and_update(state, mode)

    def prepare_fock_state(self, n, mode):
        """
        Traces out the state in 'mode' and replaces it with a Fock state defined by n.
        """
        if self._valid_modes(mode):
            with self._graph.as_default():
                n = self._maybe_batch(n, convert_to_tensor=False)
                fock_state = ops.fock_state(
                    n, D=self._cutoff_dim, pure=self._state_is_pure, batched=self._batched
                )
                self._replace_and_update(fock_state, mode)

    def prepare_coherent_state(self, alpha, mode):
        """
        Traces out the state in 'mode' and replaces it with a coherent state defined by alpha.
        """
        if self._valid_modes(mode):
            with self._graph.as_default():
                alpha = tf.cast(alpha, ops.def_type)
                alpha = self._maybe_batch(alpha)
                coherent_state = ops.coherent_state(
                    alpha, D=self._cutoff_dim, pure=self._state_is_pure, batched=self._batched
                )
                self._replace_and_update(coherent_state, mode)

    def prepare_squeezed_state(self, r, theta, mode):
        """
             Traces out the state in 'mode' and replaces it with a squeezed state defined by r and theta.
        """
        if self._valid_modes(mode):
            with self._graph.as_default():
                r = self._maybe_batch(r)
                theta = self._maybe_batch(theta)
                self._check_incompatible_batches(r, theta)
                squeezed_state = ops.squeezed_vacuum(
                    r, theta, D=self._cutoff_dim, pure=self._state_is_pure, batched=self._batched
                )
                self._replace_and_update(squeezed_state, mode)

    def prepare_displaced_squeezed_state(self, alpha, r, phi, mode):
        """
             Traces out the state in 'mode' and replaces it with a displaced squeezed state defined by alpha, r and theta.
        """
        if self._valid_modes(mode):
            with self._graph.as_default():
                alpha = self._maybe_batch(alpha)
                r = self._maybe_batch(r)
                phi = self._maybe_batch(phi)
                self._check_incompatible_batches(alpha, r, phi)
                displaced_squeezed = ops.displaced_squeezed(
                    alpha,
                    r,
                    phi,
                    D=self._cutoff_dim,
                    pure=self._state_is_pure,
                    batched=self._batched,
                )
                self._replace_and_update(displaced_squeezed, mode)

    def prepare_multimode(self, state, modes, input_state_is_pure=False):
        r"""Prepares a given mode or list of modes in the given state.

        After the preparation the system is in a mixed product state,
        with the specified modes replaced by state.

        The given state can be either in tensor form or in matrix/vector form and
        can be a batch of states or a single state. This method needs to know whether
        input_state_is_pure to distinguish between a batch of pure states and a mixed state.

        If modes is not ordered, the subsystems of the input are
        reordered to reflect that, i.e., if modes=[3,1], then the first mode
        of state ends up in mode 3 and the second mode of state ends up in
        mode 1 of the output state.

        If modes is None, it is attempted to prepare state in all modes.
        The reduced state on all other modes remains unchanged and
        the final state is product with respect to the partition into
        the modes in modes and the complement.

        Args:
            state (array): vector, matrix, or tensor representation of the ket state or
                density matrix state (or a batch of such states) in the fock basis to prepare
            modes (list[int] or non-negative int): The mode(s) into which state is
                to be prepared. Needs not be ordered.
        """
        if self._valid_modes(modes):
            if isinstance(modes, int):
                modes = [modes]

            n_modes = len(modes)
            if input_state_is_pure:
                input_is_batched = len(state.shape) > n_modes or (
                    len(state.shape) == 2 and state.shape[1] == self._cutoff_dim ** n_modes
                )
            else:
                input_is_batched = len(state.shape) % 2 == 1

            pure_shape = tuple([self._cutoff_dim] * n_modes)
            mixed_shape = tuple([self._cutoff_dim] * (2 * n_modes))
            pure_shape_as_vector = tuple([self._cutoff_dim ** n_modes])
            mixed_shape_as_matrix = tuple([self._cutoff_dim ** n_modes] * 2)
            if input_is_batched:
                pure_shape = (self._batch_size,) + pure_shape
                mixed_shape = (self._batch_size,) + mixed_shape
                pure_shape_as_vector = (self._batch_size,) + pure_shape_as_vector
                mixed_shape_as_matrix = (self._batch_size,) + mixed_shape_as_matrix

            # reshape to support input both as tensor and vector/matrix
            if state.shape == pure_shape_as_vector:
                state = tf.reshape(state, pure_shape)
            elif state.shape == mixed_shape_as_matrix:
                state = tf.reshape(state, mixed_shape)

            with self._graph.as_default():
                state = tf.cast(tf.convert_to_tensor(state), ops.def_type)
                # batch state now if not already batched and self._batched
                if self._batched and not input_is_batched:
                    state = tf.stack([state] * self._batch_size)
                self._replace_and_update(state, modes)

    def prepare_thermal_state(self, nbar, mode):
        """
        Prepares the thermal state with mean photon nbar in the specified mode.
        """
        if self._valid_modes(mode):
            with self._graph.as_default():
                nbar = self._maybe_batch(nbar)
                thermal = ops.thermal_state(nbar, D=self._cutoff_dim)
                self._replace_and_update(thermal, mode)

    def phase_shift(self, theta, mode):
        """
        Apply the phase-shift operator to the specified mode.
        """
        with self._graph.as_default():
            theta = self._maybe_batch(theta)
            new_state = ops.phase_shifter(
                theta, mode, self._state, self._cutoff_dim, self._state_is_pure, self._batched
            )
            self._update_state(new_state)

    def displacement(self, alpha, mode):
        """
        Apply the displacement operator to the specified mode.
        """
        with self._graph.as_default():
            alpha = self._maybe_batch(alpha)
            new_state = ops.displacement(
                alpha, mode, self._state, self._cutoff_dim, self._state_is_pure, self._batched
            )
            self._update_state(new_state)

    def squeeze(self, z, mode):
        """
        Apply the single-mode squeezing operator to the specified mode.
        """
        with self._graph.as_default():
            z = tf.cast(z, ops.def_type)
            r = tf.abs(z)
            x = tf.real(z)
            y = tf.imag(z)
            theta = tf.atan2(y, x)
            r = self._maybe_batch(r)
            theta = self._maybe_batch(theta)
            self._check_incompatible_batches(r, theta)
            new_state = ops.squeezer(
                r, theta, mode, self._state, self._cutoff_dim, self._state_is_pure, self._batched
            )
            self._update_state(new_state)

    def beamsplitter(self, t, r, mode1, mode2):
        """
        Apply a beamsplitter operator to the two specified modes.
        """
        with self._graph.as_default():
            t = self._maybe_batch(t)
            r = self._maybe_batch(r)
            self._check_incompatible_batches(t, r)
            new_state = ops.beamsplitter(
                t,
                r,
                mode1,
                mode2,
                self._state,
                self._cutoff_dim,
                self._state_is_pure,
                self._batched,
            )
            self._update_state(new_state)

    def kerr_interaction(self, kappa, mode):
        """
        Apply the Kerr interaction operator to the specified mode.
        """
        with self._graph.as_default():
            k = tf.cast(kappa, ops.def_type)
            k = self._maybe_batch(k)
            new_state = ops.kerr_interaction(
                k, mode, self._state, self._cutoff_dim, self._state_is_pure, self._batched
            )
            self._update_state(new_state)

    def cross_kerr_interaction(self, kappa, mode1, mode2):
        """
        Apply the cross-Kerr interaction operator to the specified mode.
        """
        with self._graph.as_default():
            k = tf.cast(kappa, ops.def_type)
            k = self._maybe_batch(k)
            new_state = ops.cross_kerr_interaction(
                k, mode1, mode2, self._state, self._cutoff_dim, self._state_is_pure, self._batched
            )
            self._update_state(new_state)

    def cubic_phase(self, gamma, mode):
        """
        Apply the cubic phase operator to the specified mode.
        """
        with self._graph.as_default():
            g = tf.cast(gamma, ops.def_type)
            g = self._maybe_batch(g)
            new_state = ops.cubic_phase(
                g,
                mode,
                self._state,
                self._cutoff_dim,
                self._hbar,
                self._state_is_pure,
                self._batched,
            )
            self._update_state(new_state)

    def loss(self, T, mode):
        """
        Apply a loss channel  to the specified mode.
        """
        with self._graph.as_default():
            # in_modes = self._state
            T = tf.cast(T, ops.def_type)
            T = self._maybe_batch(T)
            new_state = ops.loss_channel(
                T, mode, self._state, self._cutoff_dim, self._state_is_pure, self._batched
            )
            self._update_state(new_state)
            self._state_is_pure = False  # loss output always in mixed state representation

    def vacuum_element(self):
        """Compute the fidelity with the (multi-mode) vacuum state."""
        with self._graph.as_default():
            if self._batched:
                vac_component = tf.reshape(self._state, [self._batch_size, -1])[:, 0]
            else:
                vac_component = tf.reshape(self._state, [-1])[0]

            if self._state_is_pure:
                return tf.abs(vac_component) ** 2
            return vac_component

    def measure_fock(self, modes, select=None, **kwargs):
        """
        Measures 'modes' in the Fock basis and updates remaining modes conditioned on this result.
        After measurement, the states in 'modes' are reset to the vacuum.

        Args:
            modes (Sequence[int]): which modes to measure (in increasing order).
            select (Sequence[int]): user-specified measurement value (used instead of random sampling)
            **kwargs: can be used to pass a session or a feed_dict. Otherwise a temporary session
            and no feed_dict will be used.

        Returns:
            A list with the Fock number measurement results for each mode.
        """
        # allow integer (non-list) arguments
        # not part of the API, but provided for convenience
        if isinstance(modes, int):
            modes = [modes]
        if isinstance(select, int):
            select = [select]

        # convert lists to np arrays
        if isinstance(modes, list):
            modes = np.array(modes)
        if isinstance(select, list):
            select = np.array(select)

        # check for valid 'modes' argument
        if (
            len(modes) == 0 or len(modes) > self._num_modes or len(modes) != len(set(modes))
        ):  # pylint: disable=len-as-condition
            raise ValueError("Specified modes are not valid.")
        if np.any(modes != sorted(modes)):
            raise ValueError("'modes' must be sorted in increasing order.")

        # check for valid 'select' argument
        if select is not None:
            if np.any(select == None):  # pylint: disable=singleton-comparison
                raise NotImplementedError(
                    "Post-selection lists must only contain numerical values."
                )
            if self._batched:
                num_meas_modes = len(modes)
                # in this case, select must either be:
                # np array of shape (M,), or
                # np array of shape (B,M)
                # where B is the batch_size and M is the number of measured modes
                shape_err = False
                if len(select.shape) == 1:
                    # non-batched list, must broadcast
                    if select.shape[0] != num_meas_modes:
                        shape_err = True
                    else:
                        select = np.vstack([select] * self._batch_size)
                elif len(select.shape) == 2:
                    # batch of lists, no need to broadcast
                    if select.shape != (self._batch_size, num_meas_modes):
                        shape_err = True
                else:
                    shape_err = True
                if shape_err:
                    raise ValueError("The shape of 'select' is incompatible with 'modes'.")
            else:
                # in this case, select should be a vector
                if select.shape != modes.shape:
                    raise ValueError("'select' must be have the same shape as 'modes'")

        # carry out the operation
        with self.graph.as_default():
            evaluate_results, session, feed_dict, close_session = ops._check_for_eval(kwargs)
            num_reduced_state_modes = len(modes)
            reduced_state = self._state
            if self._state_is_pure:
                mode_size = 1
            else:
                mode_size = 2
            if self._batched:
                batch_size = self._batch_size
                batch_offset = 1
            else:
                batch_size = 1
                batch_offset = 0

            if select is not None:
                # just use the supplied measurement results
                meas_result = select
            else:
                # compute and sample measurement result
                if self._state_is_pure and len(modes) == self._num_modes:
                    # in this case, measure directly on the pure state
                    probs = tf.abs(self._state) ** 2
                    logprobs = tf.log(probs)
                    sample = tf.multinomial(tf.reshape(logprobs, [batch_size, -1]), 1)
                    sample_tensor = tf.squeeze(sample)
                else:
                    # otherwise, trace out unmeasured modes and sample using diagonal of reduced state
                    removed_ctr = 0
                    red_state_is_pure = self._state_is_pure
                    for m in range(self._num_modes):
                        if m not in modes:
                            new_mode_idx = m - removed_ctr
                            reduced_state = ops.partial_trace(
                                reduced_state, new_mode_idx, red_state_is_pure, self._batched
                            )
                            red_state_is_pure = False
                            removed_ctr += 1
                    # go from bra_A,ket_A,bra_B,ket_B,... -> bra_A,bra_B,ket_A,ket_B,... since this is what diag_part expects
                    # workaround for getting multi-index diagonal since tensorflow doesn't support getting diag of more than one subsystem at once
                    if num_reduced_state_modes > 1:
                        state_indices = np.arange(batch_offset + 2 * num_reduced_state_modes)
                        batch_index = state_indices[:batch_offset]
                        bra_indices = state_indices[batch_offset::2]
                        ket_indices = state_indices[batch_offset + 1 :: 2]
                        transpose_list = np.concatenate([batch_index, bra_indices, ket_indices])
                        reduced_state_reshuffled = tf.transpose(reduced_state, transpose_list)
                    else:
                        reduced_state_reshuffled = reduced_state
                    diag_indices = [self._cutoff_dim ** num_reduced_state_modes] * 2
                    if self._batched:
                        diag_indices = [self._batch_size] + diag_indices
                    diag_tensor = tf.reshape(reduced_state_reshuffled, diag_indices)
                    diag_entries = tf.matrix_diag_part(diag_tensor)
                    # hack so we can use tf.multinomial for sampling
                    logprobs = tf.log(tf.cast(diag_entries, tf.float64))
                    sample = tf.multinomial(tf.reshape(logprobs, [batch_size, -1]), 1)
                    # sample is a single integer; we need to convert it to the corresponding [n0,n1,n2,...]
                    sample_tensor = tf.squeeze(sample)

                # sample_val is a single integer for each batch entry;
                # we need to convert it to the corresponding [n0,n1,n2,...]
                meas_result = ops.unravel_index(
                    sample_tensor, [self._cutoff_dim] * num_reduced_state_modes
                )
                if not self._batched:
                    meas_result = meas_result[0]  # no batch index, can get rid of first axis

            # unstack this here because that's how it should be returned
            meas_result = tf.unstack(meas_result, axis=-1, name="Meas_result")

            # evaluate measurement result if necessary
            if evaluate_results:
                meas_result = session.run(meas_result, feed_dict=feed_dict)
                if close_session:
                    session.close()

            # project remaining modes into conditional state
            if len(modes) == self._num_modes:
                # in this case, all modes were measured and we can put everything in vacuum by reseting
                self.reset(pure=self._state_is_pure)
            else:
                # only some modes were measured: put unmeasured modes in conditional state, while reseting measured modes to vac
                fock_state = tf.one_hot(
                    tf.stack(meas_result, axis=-1), depth=self._cutoff_dim, dtype=ops.def_type
                )
                conditional_state = self._state
                for idx, mode in enumerate(modes):
                    if self._batched:
                        f = fock_state[:, idx]
                    else:
                        f = fock_state[idx]
                    conditional_state = ops.conditional_state(
                        conditional_state, f, mode, self._state_is_pure, batched=self._batched
                    )

                if self._state_is_pure:
                    norm = tf.norm(tf.reshape(conditional_state, [batch_size, -1]), axis=1)
                else:
                    # calculate norm of conditional_state
                    # use a cheap hack since tensorflow doesn't allow einsum equation for trace:
                    r = conditional_state
                    for _ in range(self._num_modes - num_reduced_state_modes - 1):
                        r = ops.partial_trace(r, 0, False, self._batched)
                    norm = tf.trace(r)

                # for broadcasting
                norm_reshape = [1] * len(conditional_state.shape[batch_offset:])
                if self._batched:
                    norm_reshape = [self._batch_size] + norm_reshape

                normalized_conditional_state = conditional_state / tf.reshape(norm, norm_reshape)

                # reset measured modes into vacuum
                single_mode_vac = (
                    self._single_mode_pure_vac
                    if self._state_is_pure
                    else self._single_mode_mixed_vac
                )
                if len(modes) == 1:
                    meas_modes_vac = single_mode_vac
                else:
                    meas_modes_vac = ops.combine_single_modes(
                        [single_mode_vac] * len(modes), self._batched
                    )
                batch_index = indices[:batch_offset]
                meas_mode_indices = indices[batch_offset : batch_offset + mode_size * len(modes)]
                conditional_indices = indices[
                    batch_offset
                    + mode_size * len(modes) : batch_offset
                    + mode_size * self._num_modes
                ]
                eqn_lhs = batch_index + meas_mode_indices + "," + batch_index + conditional_indices
                eqn_rhs = ""
                meas_ctr = 0
                cond_ctr = 0
                for m in range(self._num_modes):
                    if m in modes:
                        # use measured_indices
                        eqn_rhs += meas_mode_indices[
                            mode_size * meas_ctr : mode_size * (meas_ctr + 1)
                        ]
                        meas_ctr += 1
                    else:
                        # use conditional indices
                        eqn_rhs += conditional_indices[
                            mode_size * cond_ctr : mode_size * (cond_ctr + 1)
                        ]
                        cond_ctr += 1
                eqn = eqn_lhs + "->" + batch_index + eqn_rhs
                new_state = tf.einsum(eqn, meas_modes_vac, normalized_conditional_state)

                self._update_state(new_state)

            return tuple(meas_result)

    def measure_homodyne(self, phi, mode, select=None, **kwargs):
        """
            Measures 'modes' in the basis of quadrature eigenstates (rotated by phi)
            and updates remaining modes conditioned on this result.
            After measurement, the states in 'modes' are reset to the vacuum.

            Args:
                phi (float): phase angle of quadrature to measure
                mode (int): which mode to measure.
                select (float): user-specified measurement value (used instead of random sampling)
                **kwargs: can be used to pass a session or a feed_dict. Otherwise a temporary session
                and no feed_dict will be used.

        Returns:
            The measured value (or a list of measured values when running in batch mode).
        """

        if not isinstance(mode, int):
            raise ValueError("Specified modes are not valid.")
        else:
            if mode < 0 or mode >= self._num_modes:
                raise ValueError("Specified modes are not valid.")

        m_omega_over_hbar = 1 / self._hbar
        if self._state_is_pure:
            mode_size = 1
        else:
            mode_size = 2
        if self._batched:
            batch_offset = 1
            batch_size = self._batch_size
        else:
            batch_offset = 0
            batch_size = 1

        with self.graph.as_default():
            phi = tf.cast(phi, ops.def_type)
            phi = self._maybe_batch(phi)

            evaluate_results, session, feed_dict, close_session = ops._check_for_eval(kwargs)

            if select is not None:
                meas_result = self._maybe_batch(select)
                homodyne_sample = tf.cast(meas_result, tf.float64, name="Meas_result")
            else:
                # create reduced state on mode to be measured
                reduced_state = ops.reduced_density_matrix(
                    self._state, mode, self._state_is_pure, self._batched
                )

                # rotate to homodyne basis
                # pylint: disable=invalid-unary-operand-type
                reduced_state = ops.phase_shifter(
                    -phi, 0, reduced_state, self._cutoff_dim, False, self._batched
                )

                # create pdf for homodyne measurement
                # We use the following quadrature wavefunction for the Fock states:
                # \psi_n(x) = 1/sqrt[2^n n!](\frac{m \omega}{\pi \hbar})^{1/4}
                #             \exp{-\frac{m \omega}{2\hbar} x^2} H_n(\sqrt{\frac{m \omega}{\pi}} x)
                # where H_n(x) is the (physicists) nth Hermite polynomial
                if "max" in kwargs:
                    q_mag = kwargs["max"]
                else:
                    q_mag = 10
                if "num_bins" in kwargs:
                    num_bins = kwargs["num_bins"]
                else:
                    num_bins = 100000
                if "q_tensor" in self._cache:
                    # use cached q_tensor
                    q_tensor = self._cache["q_tensor"]
                else:
                    q_tensor = tf.constant(np.linspace(-q_mag, q_mag, num_bins))
                    self._cache["q_tensor"] = q_tensor
                x = np.sqrt(m_omega_over_hbar) * q_tensor
                if "hermite_polys" in self._cache:
                    # use cached polynomials
                    hermite_polys = self._cache["hermite_polys"]
                else:
                    H0 = 0 * x + 1.0
                    H1 = 2 * x
                    hermite_polys = [H0, H1]
                    Hn = H1
                    Hn_m1 = H0
                    for n in range(1, self._cutoff_dim - 1):
                        Hn_p1 = ops.H_n_plus_1(Hn, Hn_m1, n, x)
                        hermite_polys.append(Hn_p1)
                        Hn_m1 = Hn
                        Hn = Hn_p1
                    self._cache["hermite_polys"] = hermite_polys

                number_state_indices = [k for k in product(range(self._cutoff_dim), repeat=2)]
                terms = [
                    1
                    / np.sqrt(2 ** n * factorial(n) * 2 ** m * factorial(m))
                    * hermite_polys[n]
                    * hermite_polys[m]
                    for n, m in number_state_indices
                ]
                hermite_matrix = tf.scatter_nd(
                    number_state_indices, terms, [self._cutoff_dim, self._cutoff_dim, num_bins]
                )
                hermite_terms = tf.multiply(
                    tf.expand_dims(reduced_state, -1),
                    tf.expand_dims(tf.cast(hermite_matrix, ops.def_type), 0),
                )
                rho_dist = (
                    tf.cast(tf.reduce_sum(hermite_terms, axis=[1, 2]), tf.float64)
                    * (m_omega_over_hbar / np.pi) ** 0.5
                    * tf.exp(-(x ** 2))
                    * (q_tensor[1] - q_tensor[0])
                )  # Delta_q for normalization (only works if the bins are equally spaced)

                # use tf.multinomial to sample
                logprobs = tf.log(rho_dist)
                samples_idx = tf.multinomial(logprobs, 1)
                homodyne_sample = tf.gather(q_tensor, samples_idx)
                homodyne_sample = tf.squeeze(homodyne_sample)

            if evaluate_results:
                meas_result = homodyne_sample.eval(feed_dict, session)
                if close_session:
                    session.close()
            else:
                meas_result = tf.identity(homodyne_sample, name="Meas_result")

            # project remaining modes into conditional state
            if self._num_modes == 1:
                # in this case, all modes were measured and we we put everything into vacuum
                self.reset(pure=self._state_is_pure)
            else:
                # only some modes were measured: put unmeasured modes in conditional state, while reseting measured modes to vac
                inf_squeezed_vac = tf.convert_to_tensor(
                    [
                        (-0.5) ** (m // 2) * np.sqrt(factorial(m)) / factorial(m // 2)
                        if m % 2 == 0
                        else 0.0
                        for m in range(self._cutoff_dim)
                    ],
                    dtype=ops.def_type,
                )
                if self._batched:
                    inf_squeezed_vac = tf.tile(tf.expand_dims(inf_squeezed_vac, 0), [batch_size, 1])
                displacement_size = tf.stack(
                    tf.convert_to_tensor(meas_result * np.sqrt(m_omega_over_hbar / 2))
                )
                quad_eigenstate = ops.displacement(
                    displacement_size, 0, inf_squeezed_vac, self._cutoff_dim, True, self._batched
                )
                homodyne_eigenstate = ops.phase_shifter(
                    phi, 0, quad_eigenstate, self._cutoff_dim, True, self._batched
                )

                conditional_state = ops.conditional_state(
                    self._state,
                    homodyne_eigenstate,
                    mode,
                    self._state_is_pure,
                    batched=self._batched,
                )

                # normalize
                if self._state_is_pure:
                    norm = tf.norm(tf.reshape(conditional_state, [batch_size, -1]), axis=1)
                else:
                    # calculate norm of conditional_state
                    # cheap hack since tensorflow doesn't allow einsum equation for trace:
                    r = conditional_state
                    for _ in range(self._num_modes - 2):
                        r = ops.partial_trace(r, 0, False, self._batched)
                    norm = tf.trace(r)

                # for broadcasting
                norm_reshape = [1] * len(conditional_state.shape[batch_offset:])
                if self._batched:
                    norm_reshape = [self._batch_size] + norm_reshape

                normalized_conditional_state = conditional_state / tf.reshape(norm, norm_reshape)

                # reset measured modes into vacuum
                meas_mode_vac = (
                    self._single_mode_pure_vac
                    if self._state_is_pure
                    else self._single_mode_mixed_vac
                )
                batch_index = indices[:batch_offset]
                meas_mode_indices = indices[batch_offset : batch_offset + mode_size]
                conditional_indices = indices[
                    batch_offset + mode_size : batch_offset + mode_size * self._num_modes
                ]
                eqn_lhs = batch_index + meas_mode_indices + "," + batch_index + conditional_indices
                eqn_rhs = ""
                meas_ctr = 0
                cond_ctr = 0
                for m in range(self._num_modes):
                    if m == mode:
                        # use measured_indices
                        eqn_rhs += meas_mode_indices[
                            mode_size * meas_ctr : mode_size * (meas_ctr + 1)
                        ]
                        meas_ctr += 1
                    else:
                        # use conditional indices
                        eqn_rhs += conditional_indices[
                            mode_size * cond_ctr : mode_size * (cond_ctr + 1)
                        ]
                        cond_ctr += 1
                eqn = eqn_lhs + "->" + batch_index + eqn_rhs
                new_state = tf.einsum(eqn, meas_mode_vac, normalized_conditional_state)

                self._update_state(new_state)

            return meas_result

    @property
    def num_modes(self):
        """Number of modes in the circuit"""
        return self._num_modes

    @property
    def cutoff_dim(self):
        """Circuit cutoff dimension"""
        return self._cutoff_dim

    @property
    def state_is_pure(self):
        """Returns true if the circuit state is pure"""
        return self._state_is_pure

    @property
    def hbar(self):
        """Returns the value of hbar circuit is initialised with"""
        return self._hbar

    @property
    def graph(self):
        """Returns the computational graph"""
        return self._graph

    @property
    def batched(self):
        """Returns True if the circuit is batched"""
        return self._batched

    @property
    def batch_size(self):
        """Returns the batch size"""
        return self._batch_size

    @property
    def state(self):
        """Returns the circuit state"""
        return tf.identity(self._state, name="State")



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *

# initialize engine and program objects
eng = sf.Engine(backend="gaussian")
gate_teleportation = sf.Program(4)

with gate_teleportation.context as q:
    # create initial states
    Squeezed(0.1) | q[0]
    Squeezed(-2)  | q[1]
    Squeezed(-2)  | q[2]

    # apply the gate to be teleported
    Pgate(0.5) | q[1]

    # conditional phase entanglement
    CZgate(1) | (q[0], q[1])
    CZgate(1) | (q[1], q[2])

    # projective measurement onto
    # the position quadrature
    Fourier.H | q[0]
    MeasureX | q[0]
    Fourier.H | q[1]
    MeasureX | q[1]
    # compare against the expected output
    # X(q1/sqrt(2)).F.P(0.5).X(q0/sqrt(0.5)).F.|z>
    # not including the corrections
    Squeezed(0.1) | q[3]
    Fourier       | q[3]
    Xgate(q[0])   | q[3]
    Pgate(0.5)    | q[3]
    Fourier       | q[3]
    Xgate(q[1])   | q[3]
    # end circuit

results = eng.run(gate_teleportation)
print(results.state.reduced_gaussian([2]))
print(results.state.reduced_gaussian([3]))



#!/usr/bin/env python3
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import strawberryfields as sf
from strawberryfields.ops import *
from strawberryfields.utils import operation


@operation(4)
def prepare_squeezing(q):
    """This operation prepares modes 0 and 1
    as squeezed states with r = -1.

    Args:
        q (register): the qumode register.
    """
    S = Sgate(-1)
    S | q[0]
    S | q[1]


@operation(3)
def circuit_op(v1, v2, q):
    """Some gates that are groups into a custom operation.

    Args:
        v1 (float): parameter for CZgate
        v2 (float): parameter for the cubic phase gate
        q (register): the qumode register
    """
    CZgate(v1) | (q[0], q[1])
    Vgate(v2) | q[1]


# initialize engine and program objects
eng = sf.Engine(backend="fock", backend_options={"cutoff_dim": 5})
circuit = sf.Program(4)

with circuit.context as q:
    # The following operation takes no arguments
    prepare_squeezing() | q

    # another operation with 2 parameters that operates on three registers: 0, 1
    circuit_op(0.5719, 2.0603) | (q[0], q[1], q[3])

# run the engine
results = eng.run(circuit)



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Circuit class specification for the X12 class of circuits."""
import textwrap

import numpy as np
from numpy.linalg import multi_dot
from scipy.linalg import block_diag

from strawberryfields.decompositions import mach_zehnder
from strawberryfields.program_utils import CircuitError, Command, group_operations
from strawberryfields.parameters import par_evaluate
import strawberryfields.ops as ops

from .circuit_specs import CircuitSpecs
from .gbs import GBSSpecs


# Supporting multiple string formatting, such that the target can be replaced
# first followed by squeezing amplitude and phase values
X12_CIRCUIT = textwrap.dedent(
    """\
    name template_6x2_X12
    version 1.0
    target {target} (shots=1)

    # for n spatial degrees, first n signal modes, then n idler modes, all phases zero
    S2gate({{squeezing_amplitude_0}}, 0.0) | [0, 6]
    S2gate({{squeezing_amplitude_1}}, 0.0) | [1, 7]
    S2gate({{squeezing_amplitude_2}}, 0.0) | [2, 8]
    S2gate({{squeezing_amplitude_3}}, 0.0) | [3, 9]
    S2gate({{squeezing_amplitude_4}}, 0.0) | [4, 10]
    S2gate({{squeezing_amplitude_5}}, 0.0) | [5, 11]

    # standard 6x6 interferometer for the signal modes (the lower ones in frequency)
    # even phase indices correspond to internal Mach-Zehnder interferometer phases
    # odd phase indices correspond to external Mach-Zehnder interferometer phases
    MZgate({{phase_0}}, {{phase_1}}) | [0, 1]
    MZgate({{phase_2}}, {{phase_3}}) | [2, 3]
    MZgate({{phase_4}}, {{phase_5}}) | [4, 5]
    MZgate({{phase_6}}, {{phase_7}}) | [1, 2]
    MZgate({{phase_8}}, {{phase_9}}) | [3, 4]

    MZgate({{phase_10}}, {{phase_11}}) | [0, 1]

    MZgate({{phase_12}}, {{phase_13}}) | [2, 3]
    MZgate({{phase_14}}, {{phase_15}}) | [4, 5]
    MZgate({{phase_16}}, {{phase_17}}) | [1, 2]
    MZgate({{phase_18}}, {{phase_19}}) | [3, 4]

    MZgate({{phase_20}}, {{phase_21}}) | [0, 1]
    MZgate({{phase_22}}, {{phase_23}}) | [2, 3]
    MZgate({{phase_24}}, {{phase_25}}) | [4, 5]
    MZgate({{phase_26}}, {{phase_27}}) | [1, 2]
    MZgate({{phase_28}}, {{phase_29}}) | [3, 4]

    # duplicate the interferometer for the idler modes (the higher ones in frequency)
    MZgate({{phase_0}}, {{phase_1}}) | [6, 7]
    MZgate({{phase_2}}, {{phase_3}}) | [8, 9]
    MZgate({{phase_4}}, {{phase_5}}) | [10, 11]
    MZgate({{phase_6}}, {{phase_7}}) | [7, 8]
    MZgate({{phase_8}}, {{phase_9}}) | [9, 10]

    MZgate({{phase_10}}, {{phase_11}}) | [6, 7]
    MZgate({{phase_12}}, {{phase_13}}) | [8, 9]
    MZgate({{phase_14}}, {{phase_15}}) | [10, 11]
    MZgate({{phase_16}}, {{phase_17}}) | [7, 8]
    MZgate({{phase_18}}, {{phase_19}}) | [9, 10]

    MZgate({{phase_20}}, {{phase_21}}) | [6, 7]
    MZgate({{phase_22}}, {{phase_23}}) | [8, 9]
    MZgate({{phase_24}}, {{phase_25}}) | [10, 11]
    MZgate({{phase_26}}, {{phase_27}}) | [7, 8]
    MZgate({{phase_28}}, {{phase_29}}) | [9, 10]

    # add final dummy phases to allow mapping any unitary to this template (these do not
    # affect the photon number measurement)
    Rgate({{final_phase_0}}) | [0]
    Rgate({{final_phase_1}}) | [1]
    Rgate({{final_phase_2}}) | [2]
    Rgate({{final_phase_3}}) | [3]
    Rgate({{final_phase_4}}) | [4]
    Rgate({{final_phase_5}}) | [5]
    Rgate({{final_phase_6}}) | [6]
    Rgate({{final_phase_7}}) | [7]
    Rgate({{final_phase_8}}) | [8]
    Rgate({{final_phase_9}}) | [9]
    Rgate({{final_phase_10}}) | [10]
    Rgate({{final_phase_11}}) | [11]

    # measurement in Fock basis
    MeasureFock() | [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    """
)


class X12Specs(CircuitSpecs):
    """Circuit specifications for the X12 class of circuits."""

    short_name = "X12"
    modes = 12
    remote = True
    local = True
    interactive = False
    circuit = X12_CIRCUIT.format(target=short_name)

    sq_amplitude = 1.0

    primitives = {"S2gate", "MeasureFock", "Rgate", "BSgate", "MZgate"}
    decompositions = {
        "Interferometer": {"mesh": "rectangular_symmetric", "drop_identity": False},
        "BipartiteGraphEmbed": {"mesh": "rectangular_symmetric", "drop_identity": False},
    }

    def compile(self, seq, registers):
        """Try to arrange a quantum circuit into a form suitable for X12.

        Args:
            seq (Sequence[Command]): quantum circuit to modify
            registers (Sequence[RegRefs]): quantum registers
        Returns:
            List[Command]: modified circuit
        Raises:
            CircuitError: the circuit does not correspond to X12
        """
        # pylint: disable=too-many-statements,too-many-branches

        # first do general GBS compilation to make sure
        # Fock measurements are correct
        # ---------------------------------------------
        seq = GBSSpecs().compile(seq, registers)
        A, B, C = group_operations(seq, lambda x: isinstance(x, ops.MeasureFock))

        if len(B[0].reg) != self.modes:
            raise CircuitError("All modes must be measured.")

        # Check circuit begins with two mode squeezers
        # --------------------------------------------
        A, B, C = group_operations(seq, lambda x: isinstance(x, ops.S2gate))

        regrefs = set()

        if B:
            # get set of circuit registers as a tuple for each S2gate
            regrefs = {(cmd.reg[0].ind, cmd.reg[1].ind) for cmd in B}

        # the set of allowed mode-tuples the S2gates must have
        allowed_modes = set(zip(range(0, 6), range(6, 12)))

        if not regrefs.issubset(allowed_modes):
            raise CircuitError("S2gates do not appear on the correct modes.")

        # ensure provided S2gates all have the allowed squeezing values
        allowed_sq_value = {(0.0, 0.0), (self.sq_amplitude, 0.0)}
        sq_params = {(float(np.round(cmd.op.p[0], 3)), float(cmd.op.p[1])) for cmd in B}

        if not sq_params.issubset(allowed_sq_value):
            wrong_params = sq_params - allowed_sq_value
            raise CircuitError(
                "Incorrect squeezing value(s) (r, phi)={}. Allowed squeezing "
                "value(s) are (r, phi)={}.".format(wrong_params, allowed_sq_value)
            )

        # determine which modes do not have input S2gates specified
        missing = allowed_modes - regrefs

        for i, j in missing:
            # insert S2gates with 0 squeezing
            seq.insert(0, Command(ops.S2gate(0, 0), [registers[i], registers[j]]))

        # Check if matches the circuit template
        # --------------------------------------------
        # This will avoid superfluous unitary compilation.
        try:
            seq = super().compile(seq, registers)
        except CircuitError:
            # failed topology check. Continue to more general
            # compilation below.
            pass
        else:
            return seq

        # Compile the unitary: combine and then decompose all unitaries
        # -------------------------------------------------------------
        A, B, C = group_operations(
            seq, lambda x: isinstance(x, (ops.Rgate, ops.BSgate, ops.MZgate))
        )

        # begin unitary lists for mode [0, 1, 2, 3, 4, 5] and modes [6, 7, 8, 9, 10, 11]
        # with two identity matrices. This is because multi_dot requires at
        # least two matrices in the list.
        U_list0 = [np.identity(self.modes // 2, dtype=np.complex128)] * 2
        U_list6 = [np.identity(self.modes // 2, dtype=np.complex128)] * 2

        if not B:
            # no interferometer was applied
            A, B, C = group_operations(seq, lambda x: isinstance(x, ops.S2gate))
            A = B  # move the S2gates to A
        else:
            for cmd in B:
                # calculate the unitary matrix representing each
                # rotation gate and each beamsplitter
                modes = [i.ind for i in cmd.reg]
                params = par_evaluate(cmd.op.p)
                U = np.identity(self.modes // 2, dtype=np.complex128)

                if isinstance(cmd.op, ops.Rgate):
                    m = modes[0]
                    U[m % 6, m % 6] = np.exp(1j * params[0])

                elif isinstance(cmd.op, ops.MZgate):
                    m, n = modes
                    U = mach_zehnder(m % 6, n % 6, params[0], params[1], self.modes // 2)

                elif isinstance(cmd.op, ops.BSgate):
                    m, n = modes

                    t = np.cos(params[0])
                    r = np.exp(1j * params[1]) * np.sin(params[0])

                    U[m % 6, m % 6] = t
                    U[m % 6, n % 6] = -np.conj(r)
                    U[n % 6, m % 6] = r
                    U[n % 6, n % 6] = t

                if set(modes).issubset({0, 1, 2, 3, 4, 5}):
                    U_list0.insert(0, U)
                elif set(modes).issubset({6, 7, 8, 9, 10, 11}):
                    U_list6.insert(0, U)
                else:
                    raise CircuitError(
                        "Unitary must be applied separately to modes [0, 1, 2, 3, 4, 5] and modes [6, 7, 8, 9, 10, 11]."
                    )

        # multiply all unitaries together, to get the final
        # unitary representation on modes [0, 1] and [2, 3].
        U0 = multi_dot(U_list0)
        U6 = multi_dot(U_list6)

        # check unitaries are equal
        if not np.allclose(U0, U6):
            raise CircuitError(
                "Interferometer on modes [0, 1, 2, 3, 4, 5] must be identical to interferometer on modes [6, 7, 8, 9, 10, 11]."
            )

        U = block_diag(U0, U6)

        # replace B with an interferometer
        B = [
            Command(ops.Interferometer(U0), registers[:6]),
            Command(ops.Interferometer(U6), registers[6:]),
        ]

        # decompose the interferometer, using Mach-Zehnder interferometers
        B = self.decompose(B)

        # Do a final circuit topology check
        # ---------------------------------
        seq = super().compile(A + B + C, registers)
        return seq


class X12_01(X12Specs):
    """Circuit specifications for the first X12 chip."""

    short_name = "X12_01"
    circuit = X12_CIRCUIT.format(target=short_name)


class X12_02(X12Specs):
    """Circuit specifications for the second X12 chip."""

    short_name = "X12_02"
    circuit = X12_CIRCUIT.format(target=short_name)



# Copyright 2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This package contains the modules for the low-level Strawberry Fields program
execution API. The :class:`~strawberryfields.api.Connection` class mediates
the network connection to, and exposes operations provided by, a remote program
execution backend. The :class:`~strawberryfields.api.Job` and
:class:`~strawberryfields.api.Result` classes provide interfaces for managing
program execution jobs and job results respectively.
"""

from .connection import Connection, RequestFailedError
from .job import InvalidJobOperationError, Job, JobStatus
from .result import Result

__all__ = ["Connection", "Job", "Result"]
